# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='phd',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x0c\x63onfig.proto\x12\x03phd\"\xc6\x01\n\x0cGlobalConfig\x12\x14\n\x05uname\x18\x01 \x02(\tR\x05uname\x12!\n\x0c\x63onfigure_id\x18\x02 \x02(\tR\x0b\x63onfigureId\x12\x1b\n\twith_cuda\x18\x64 \x02(\x08R\x08withCuda\x12\x32\n\x07options\x18\x03 \x02(\x0b\x32\x18.phd.GlobalConfigOptionsR\x07options\x12,\n\x05paths\x18\x04 \x02(\x0b\x32\x16.phd.GlobalConfigPathsR\x05paths\"\xb9\x01\n\x13GlobalConfigOptions\x12\x1b\n\twith_cuda\x18\x01 \x02(\x08R\x08withCuda\x12\x32\n\x15update_git_submodules\x18\x02 \x02(\x08R\x13updateGitSubmodules\x12%\n\x0esymlink_python\x18\x03 \x02(\x08R\rsymlinkPython\x12*\n\x11install_git_hooks\x18\x04 \x02(\x08R\x0finstallGitHooks\"H\n\x11GlobalConfigPaths\x12\x1b\n\trepo_root\x18\x01 \x02(\tR\x08repoRoot\x12\x16\n\x06python\x18\x02 \x02(\tR\x06python\"\xf9\x02\n\tBuildInfo\x12.\n\x13seconds_since_epoch\x18\x01 \x02(\x03R\x11secondsSinceEpoch\x12\x12\n\x04host\x18\x02 \x02(\tR\x04host\x12\x12\n\x04user\x18\x03 \x02(\tR\x04user\x12\x1d\n\nrepo_dirty\x18\x05 \x02(\x08R\trepoDirty\x12\x1d\n\ngit_commit\x18\x04 \x02(\tR\tgitCommit\x12\x1d\n\ngit_branch\x18\x06 \x01(\tR\tgitBranch\x12!\n\x0cgit_tracking\x18\x07 \x01(\tR\x0bgitTracking\x12$\n\x0egit_remote_url\x18\x08 \x01(\tR\x0cgitRemoteUrl\x12*\n\x11git_commit_author\x18\n \x01(\tR\x0fgitCommitAuthor\x12\x42\n\x1egit_commit_seconds_since_epoch\x18\x0b \x01(\x03R\x1agitCommitSecondsSinceEpoch')
)




_GLOBALCONFIG = _descriptor.Descriptor(
  name='GlobalConfig',
  full_name='phd.GlobalConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uname', full_name='phd.GlobalConfig.uname', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='uname', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='configure_id', full_name='phd.GlobalConfig.configure_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='configureId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='with_cuda', full_name='phd.GlobalConfig.with_cuda', index=2,
      number=100, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='withCuda', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='options', full_name='phd.GlobalConfig.options', index=3,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='options', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='paths', full_name='phd.GlobalConfig.paths', index=4,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='paths', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=220,
)


_GLOBALCONFIGOPTIONS = _descriptor.Descriptor(
  name='GlobalConfigOptions',
  full_name='phd.GlobalConfigOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='with_cuda', full_name='phd.GlobalConfigOptions.with_cuda', index=0,
      number=1, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='withCuda', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='update_git_submodules', full_name='phd.GlobalConfigOptions.update_git_submodules', index=1,
      number=2, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='updateGitSubmodules', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symlink_python', full_name='phd.GlobalConfigOptions.symlink_python', index=2,
      number=3, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='symlinkPython', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='install_git_hooks', full_name='phd.GlobalConfigOptions.install_git_hooks', index=3,
      number=4, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='installGitHooks', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=223,
  serialized_end=408,
)


_GLOBALCONFIGPATHS = _descriptor.Descriptor(
  name='GlobalConfigPaths',
  full_name='phd.GlobalConfigPaths',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='repo_root', full_name='phd.GlobalConfigPaths.repo_root', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='repoRoot', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='python', full_name='phd.GlobalConfigPaths.python', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='python', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=410,
  serialized_end=482,
)


_BUILDINFO = _descriptor.Descriptor(
  name='BuildInfo',
  full_name='phd.BuildInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='seconds_since_epoch', full_name='phd.BuildInfo.seconds_since_epoch', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='secondsSinceEpoch', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='host', full_name='phd.BuildInfo.host', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='host', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user', full_name='phd.BuildInfo.user', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='user', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='repo_dirty', full_name='phd.BuildInfo.repo_dirty', index=3,
      number=5, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='repoDirty', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_commit', full_name='phd.BuildInfo.git_commit', index=4,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitCommit', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_branch', full_name='phd.BuildInfo.git_branch', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitBranch', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_tracking', full_name='phd.BuildInfo.git_tracking', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitTracking', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_remote_url', full_name='phd.BuildInfo.git_remote_url', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitRemoteUrl', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_commit_author', full_name='phd.BuildInfo.git_commit_author', index=8,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitCommitAuthor', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='git_commit_seconds_since_epoch', full_name='phd.BuildInfo.git_commit_seconds_since_epoch', index=9,
      number=11, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='gitCommitSecondsSinceEpoch', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=485,
  serialized_end=862,
)

_GLOBALCONFIG.fields_by_name['options'].message_type = _GLOBALCONFIGOPTIONS
_GLOBALCONFIG.fields_by_name['paths'].message_type = _GLOBALCONFIGPATHS
DESCRIPTOR.message_types_by_name['GlobalConfig'] = _GLOBALCONFIG
DESCRIPTOR.message_types_by_name['GlobalConfigOptions'] = _GLOBALCONFIGOPTIONS
DESCRIPTOR.message_types_by_name['GlobalConfigPaths'] = _GLOBALCONFIGPATHS
DESCRIPTOR.message_types_by_name['BuildInfo'] = _BUILDINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GlobalConfig = _reflection.GeneratedProtocolMessageType('GlobalConfig', (_message.Message,), dict(
  DESCRIPTOR = _GLOBALCONFIG,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.GlobalConfig)
  ))
_sym_db.RegisterMessage(GlobalConfig)

GlobalConfigOptions = _reflection.GeneratedProtocolMessageType('GlobalConfigOptions', (_message.Message,), dict(
  DESCRIPTOR = _GLOBALCONFIGOPTIONS,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.GlobalConfigOptions)
  ))
_sym_db.RegisterMessage(GlobalConfigOptions)

GlobalConfigPaths = _reflection.GeneratedProtocolMessageType('GlobalConfigPaths', (_message.Message,), dict(
  DESCRIPTOR = _GLOBALCONFIGPATHS,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.GlobalConfigPaths)
  ))
_sym_db.RegisterMessage(GlobalConfigPaths)

BuildInfo = _reflection.GeneratedProtocolMessageType('BuildInfo', (_message.Message,), dict(
  DESCRIPTOR = _BUILDINFO,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:phd.BuildInfo)
  ))
_sym_db.RegisterMessage(BuildInfo)


# @@protoc_insertion_point(module_scope)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilies for grokking the underlying system.

Variables:
  * `HOSTNAME` (str) System hostname.
  * `USERNAME` (str) Username.
  * `UID` (int) User ID.
  * `PID` (int) Process ID.
"""
from __future__ import print_function

import getpass
import os
import socket
import subprocess
import sys
import tempfile
import threading
import typing
from sys import platform

from labm8 import app
from labm8 import fs

HOSTNAME = socket.gethostname()
USERNAME = getpass.getuser()
UID = os.getuid()
PID = os.getpid()

argv = sys.argv
STDOUT = sys.stdout
STDERR = sys.stderr
PIPE = subprocess.PIPE


class Error(Exception):
  pass


class SubprocessError(Error):
  """
  Error thrown if a subprocess fails.
  """
  pass


class CommandNotFoundError(Exception):
  """
  Error thrown a system command is not found.
  """
  pass


class ScpError(Error):
  """
  Error thrown if scp file transfer fails.
  """

  def __init__(self, stdout, stderr):
    """
    Construct an ScpError.

    Arguments:

        stdout (str): Captured stdout of scp subprocess.
        stderr (str): Captured stderr of scp subprocess.
    """
    self.out = stdout
    self.err = stderr

  def __repr__(self):
    return self.out + '\n' + self.err

  def __str__(self):
    return self.__repr__()


class Subprocess(object):
  """Subprocess abstraction.

  Wrapper around subprocess.Popen() which provides the ability to
  force a timeout after a number of seconds have elapsed.
  """

  def __init__(
      self,
      cmd,
      shell=False,
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
      decode_out=True,
  ):
    """
    Create a new subprocess.
    """
    self.cmd = cmd
    self.process = None
    self.stdout = None
    self.stderr = None
    self.shell = shell
    self.decode_out = decode_out

    self.stdout_dest = stdout
    self.stderr_dest = stderr

  def run(self, timeout=-1):
    """
    Run the subprocess.

    Arguments:
        timeout (optional) If a positive real value, then timout after
            the given number of seconds.

    Raises:
        SubprocessError If subprocess has not completed after "timeout"
            seconds.
    """

    def target():
      self.process = subprocess.Popen(
          self.cmd,
          stdout=self.stdout_dest,
          stderr=self.stderr_dest,
          shell=self.shell,
      )
      stdout, stderr = self.process.communicate()

      # Decode output if the user wants, and if there is any.
      if self.decode_out:
        if stdout:
          self.stdout = stdout.decode('utf-8')
        if stderr:
          self.stderr = stderr.decode('utf-8')

    thread = threading.Thread(target=target)
    thread.start()

    if timeout > 0:
      thread.join(timeout)
      if thread.is_alive():
        self.process.terminate()
        thread.join()
        raise SubprocessError(
            ('Reached timeout after {t} seconds'.format(t=timeout)),)
    else:
      thread.join()

    return self.process.returncode, self.stdout, self.stderr


def is_linux():
  return platform == 'linux' or platform == 'linux2'


def is_mac():
  return platform == 'darwin'


def is_windows():
  return platform == 'win32'


def run(command, num_retries=1, timeout=-1, **kwargs):
  """
  Run a command with optional timeout and retries.

  Provides a convenience method for executing a subprocess with
  additional error handling.

  Arguments:
      command (list of str): The command to execute.
      num_retries (int, optional): If the subprocess fails, the number of
        attempts to execute it before failing.
      timeout (float, optional): If positive, the number of seconds to wait
        for subprocess completion before failing.
      **kwargs: Additional args to pass to Subprocess.__init__()

  Returns:
      Tuple of (int, str, str): Where the variables represent
      (exit status, stdout, stderr).

  Raises:
      SubprocessError: If the command fails after the given number of
        retries.
  """
  last_error = None
  for _ in range(num_retries):
    try:
      process = Subprocess(command, **kwargs)
      return process.run(timeout)
    except Exception as err:
      last_error = err

  raise last_error


def sed(match, replacement, path, modifiers=''):
  """Perform sed text substitution.

  This requires GNU sed. On MacOS, install it using:

      $ brew "gnu-sed"

  And then ensure that it is in the PATH before the OS-shipped sed:

      $ export PATH="/usr/local/opt/gnu-sed/libexec/gnubin:$PATH"
  """
  cmd = "sed -r -i 's/%s/%s/%s' %s" % (match, replacement, modifiers, path)

  process = Subprocess(cmd, shell=True)
  ret, out, err = process.run(timeout=60)
  if ret:
    raise SubprocessError('Sed command failed!')


def echo(*args, **kwargs):
  """
  Write a message to a file.

  Arguments:
      args A list of arguments which make up the message. The last argument
          is the path to the file to write to.
  """
  msg = args[:-1]
  path = fs.path(args[-1])
  append = kwargs.pop('append', False)

  if append:
    with open(path, 'a') as file:
      print(*msg, file=file, **kwargs)
  else:
    with open(fs.path(path), 'w') as file:
      print(*msg, file=file, **kwargs)


def which(program, path=None):
  """
  Returns the full path of shell commands.

  Replicates the functionality of system which (1) command. Looks
  for the named program in the directories indicated in the $PATH
  environment variable, and returns the full path if found.

  Examples:

      >>> system.which("ls")
      "/bin/ls"

      >>> system.which("/bin/ls")
      "/bin/ls"

      >>> system.which("not-a-real-command")
      None

      >>> system.which("ls", path=("/usr/bin", "/bin"))
      "/bin/ls"

  Arguments:

      program (str): The name of the program to look for. Can
        be an absolute path.
      path (sequence of str, optional): A list of directories to
        look for the pgoram in. Default value is system $PATH.

  Returns:

     str: Full path to program if found, else None.
  """
  # If path is not given, read the $PATH environment variable.
  path = path or os.environ['PATH'].split(os.pathsep)
  abspath = True if os.path.split(program)[0] else False
  if abspath:
    if fs.isexe(program):
      return program
  else:
    for directory in path:
      # De-quote directories.
      directory = directory.strip('"')
      exe_file = os.path.join(directory, program)
      if fs.isexe(exe_file):
        return exe_file

  return None


def isprocess(pid, error=False):
  """
  Check that a process is running.

  Arguments:

      pid (int): Process ID to check.

  Returns:

      True if the process is running, else false.
  """
  try:
    # Don't worry folks, no processes are harmed in the making of
    # this system call:
    os.kill(pid, 0)
    return True
  except OSError:
    return False


def exit(status=0):
  """
  Terminate the program with the given status code.
  """
  if status == 0:
    print('Done.', file=sys.stderr)
  else:
    print('Error {0}'.format(status), file=sys.stderr)
  sys.exit(status)


def ProcessFileAndReplace(
    path: str,
    process_file_callback: typing.Callable[[str, str], None],
    tempfile_prefix: str = 'labm8_system_',
    tempfile_suffix: str = None,
) -> None:
  """Process a file and replace with the generated file.

  This function provides the functionality of inplace file modification for
  functions which take an input file and produce an output file. It does this
  by creating a temporary file which, if the function returns successfully (i.e.
  without exception), will overwrite the original file.

  Args:
    path: The path of the file to process inplace.
    process_file_callback: A function which takes two arguments - the path of
      an input file, and the path of an output file.
    tempfile_prefix: An optional name prefix for the temporary file.
    tempfile_suffix: An optional name suffix for the temporary file.
  """
  with tempfile.NamedTemporaryFile(
      prefix=tempfile_prefix,
      suffix=tempfile_suffix,
      delete=False,
  ) as f:
    tmp_path = f.name
    try:
      process_file_callback(path, tmp_path)
      os.rename(tmp_path, path)
    finally:
      if os.path.isfile(tmp_path):
        os.unlink(tmp_path)


def CheckCallOrDie(cmd: typing.List[str]) -> None:
  """Run the given command and exit fatally on error."""
  try:
    app.Log(2, '$ %s', ' '.join(cmd))
    subprocess.check_call(cmd)
  except subprocess.CalledProcessError as e:
    app.FatalWithoutStackTrace(
        'Command: `%s` failed with error: %s',
        ' '.join(cmd),
        e,
    )



"""Module for handling archive files."""
import pathlib
import shutil
import tarfile
import tempfile
import typing
import zipfile


class UnsupportedArchiveFormat(ValueError):
  """Raised in case an archive has an unsupported file format."""
  pass


class Archive(object):
  """An archive file.

  Provides uniform access unpacked archives when used as a context manager by
  extracting the archive contents to a temporary directory.

  Example:
    >>> with Archive("/tmp/data.zip") as uncompressed_root:
    ...   print(uncompressed_root.iterdir())
    ['a', 'README.txt']

  If the archive is a bazel data dependency, you can use the subclass
  labm8.bazelutil.DataArchive to resolve the absolute path.
  """

  def __init__(
      self,
      path: typing.Union[str, pathlib.Path],
      assume_filename: typing.Optional[typing.Union[str, pathlib.Path]] = None,
  ):
    """Create an archive.

    Will determine the type of the archive from the suffix, e.g. if path is
    'foo.zip', will treat the file as a zip file. The assume_filename path
    can be used to change the determined type.

    Args:
      path: The path to the data, including the name of the workspace.
      assume_filename: For the purpose of determining the encoding of the
        archive from the file extension, use this name rather than the true
        path.

    Raises:
      FileNotFoundError: If path is not a file.
    """
    self._compressed_path = pathlib.Path(path)
    if not self._compressed_path.is_file():
      raise FileNotFoundError(f"No such file: '{path}'")

    # The path used to determine the type of the archive.
    path_to_determine_type = pathlib.Path(assume_filename or path)
    suffixes = path_to_determine_type.suffixes

    if not suffixes:
      raise UnsupportedArchiveFormat(
          f"Archive '{path_to_determine_type.name}' has no extension",)

    if suffixes[-1] == '.zip':
      self._open_function = zipfile.ZipFile
    elif suffixes[-2:] == ['.tar', '.bz2']:
      self._open_function = lambda f: tarfile.open(f, 'r:bz2')
      # TODO(cec): Add support for .tar, and .tar.gz.
    else:
      raise UnsupportedArchiveFormat(
          f"Unsupported file extension '{suffixes[-1]}' for archive "
          f"'{path_to_determine_type.name}'",)

    # Set in __enter__().
    self._uncompressed_path: typing.Optional[pathlib.Path] = None

  @property
  def path(self) -> pathlib.Path:
    """Return the path of the archive."""
    return self._compressed_path

  def ExtractAll(self, path: pathlib.Path) -> pathlib.Path:
    """Extract the archive contents to a directory.

    Args:
      path: The directory to extract to.

    Returns:
      The path of the extracted archive.
    """
    with self._open_function(str(self._compressed_path)) as f:
      f.extractall(path=str(path))
    return path

  def __enter__(self) -> pathlib.Path:
    """Unpack the archive and return the uncompressed path.

    Returns:
      The path of the directory containing the uncompressed archive.
    """
    assert not self._uncompressed_path
    self._uncompressed_path = pathlib.Path(tempfile.mkdtemp(prefix='phd_'))
    return self.ExtractAll(self._uncompressed_path)

  def __exit__(self, *args):
    """Exit the scope of the archive.

    This deletes the temporary directory that the archive has been unpacked to.
    """
    assert self._uncompressed_path
    shutil.rmtree(self._uncompressed_path)
    self._uncompressed_path = None



"""Internal logging library implementation."""
import fnmatch
import functools
import sys

from absl import flags as absl_flags
from absl import logging as absl_logging

FLAGS = absl_flags.FLAGS

absl_flags.DEFINE_list(
    'vmodule', [],
    'Per-module verbose level. The argument has to contain a comma-separated '
    'list of <module name>=<log level>. <module name> is a glob pattern (e.g., '
    "gfs* for all modules whose name starts with \"gfs\"), matched against the "
    'filename base (that is, name ignoring .py). <log level> overrides any '
    'value given by --v.')

# Logging functions.

# This is a set of module ids for the modules that disclaim key flags.
# This module is explicitly added to this set so that we never consider it to
# define key flag.
disclaim_module_ids = set([id(sys.modules[__name__])])


def get_module_object_and_name(globals_dict):
  """Returns the module that defines a global environment, and its name.
  Args:
    globals_dict: A dictionary that should correspond to an environment
      providing the values of the globals.
  Returns:
    _ModuleObjectAndName - pair of module object & module name.
    Returns (None, None) if the module could not be identified.
  """
  name = globals_dict.get('__name__', None)
  module = sys.modules.get(name, None)
  # Pick a more informative name for the main module.
  return module, (sys.argv[0] if name == '__main__' else name)


def GetCallingModuleName():
  """Returns the module that's calling into this module.
  We generally use this function to get the name of the module calling a
  DEFINE_foo... function.
  Returns:
    The module name that called into this one.
  Raises:
    AssertionError: Raised when no calling module could be identified.
  """
  for depth in range(1, sys.getrecursionlimit()):
    # sys._getframe is the right thing to use here, as it's the best
    # way to walk up the call stack.
    globals_for_frame = sys._getframe(depth).f_globals  # pylint: disable=protected-access
    module, module_name = get_module_object_and_name(globals_for_frame)
    if id(module) not in disclaim_module_ids and module_name is not None:
      return module_name
  raise AssertionError('No module was found')


@functools.lru_cache(maxsize=1)
def ModuleGlob():
  return [(x.split('=')[0], int(x.split('=')[1])) for x in FLAGS.vmodule]


@functools.lru_cache(maxsize=128)
def GetModuleVerbosity(module: str) -> int:
  """Return the verbosity level for the given module."""
  module_basename = module.split('.')[-1]
  for module_glob, level in ModuleGlob():
    if fnmatch.fnmatch(module_basename, module_glob):
      return level

  return absl_logging.get_verbosity() + 1


# Skip this function when determining the calling module and line number for
# logging.
@absl_logging.skip_log_prefix
def Log(calling_module_name: str, level: int, msg, *args, **kwargs):
  """Logs a message at the given level.

  Per-module verbose level. The argument has to contain a comma-separated
  list of <module name>=<log level>. <module name> is a glob pattern (e.g., "
    "gfs* for all modules whose name starts with \"gfs\"), matched against the "
    "filename base (that is, name ignoring .py). <log level> overrides any "
    "value given by --v."
  """
  module_level = GetModuleVerbosity(calling_module_name)
  if level <= module_level:
    absl_logging.info(msg, *args, **kwargs)


@absl_logging.skip_log_prefix
def Fatal(msg, *args, **kwargs):
  """Logs a fatal message."""
  absl_logging.fatal(msg, *args, **kwargs)


@absl_logging.skip_log_prefix
def Error(msg, *args, **kwargs):
  """Logs an error message."""
  absl_logging.error(msg, *args, **kwargs)


@absl_logging.skip_log_prefix
def Warning(msg, *args, **kwargs):
  """Logs a warning message."""
  absl_logging.warning(msg, *args, **kwargs)


def FlushLogs():
  """Flushes all log files."""
  absl_logging.flush()


def DebugLogging() -> bool:
  """Return whether debug logging is enabled."""
  return absl_logging.level_debug()


def SetLogLevel(level: int) -> None:
  """Sets the logging verbosity.

  Causes all messages of level <= v to be logged, and all messages of level > v
  to be silently discarded.

  Args:
    level: the verbosity level as an integer.
  """
  absl_logging.set_verbosity(level)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Hashing and cryptography utils.
"""
import hashlib
import pathlib
import typing


def _checksum(hash_fn, data):
  return hash_fn(data).hexdigest()


def _checksum_str(hash_fn, string, encoding='utf-8'):
  return _checksum(hash_fn, string.encode(encoding))


def _checksum_list(hash_fn, *elems):
  string = ''.join(sorted(str(x) for x in elems))
  return _checksum_str(hash_fn, string)


def _checksum_file(hash_fn, path: typing.Union[str, pathlib.Path]):
  with open(path, 'rb') as infile:
    ret = _checksum(hash_fn, infile.read())
  return ret


def sha1(data):
  """
  Return the sha1 of "data".

  Arguments:
      data (bytes): Data.

  Returns:
      str: Hex encoded.
  """
  return _checksum(hashlib.sha1, data)


def sha1_str(string, encoding='utf-8'):
  """
  Return the sha1 of string "data".

  Arguments:
      string: String.

  Returns:
      str: Hex encoded.
  """
  return _checksum_str(hashlib.sha1, string, encoding=encoding)


def sha1_list(*elems):
  """
  Return the sha1 of all elements of a list.

  Arguments:
      *elems: List of stringifiable data.

  Returns:
      str: Hex encoded.
  """
  return _checksum_list(hashlib.sha1, *elems)


def sha1_file(path: typing.Union[str, pathlib.Path]):
  """
  Return the sha1 of file at "path".

  Arguments:
      path (str): Path to file

  Returns:
      str: Hex encoded.
  """
  return _checksum_file(hashlib.sha1, path)


def md5(data):
  """
  Return the md5 of "data".

  Arguments:
      data (bytes): Data.

  Returns:
      str: Hex encoded.
  """
  return _checksum(hashlib.md5, data)


def md5_str(string, encoding='utf-8'):
  """
  Return the md5 of string "data".

  Arguments:
      string: String.

  Returns:
      str: Hex encoded.
  """
  return _checksum_str(hashlib.md5, string, encoding=encoding)


def md5_list(*elems):
  """
  Return the md5 of all elements of a list.

  Arguments:
      *elems: List of stringifiable data.

  Returns:
      str: Hex encoded.
  """
  return _checksum_list(hashlib.md5, *elems)


def md5_file(path: typing.Union[str, pathlib.Path]):
  """
  Return the md5 of file at "path".

  Arguments:
      path (str): Path to file

  Returns:
      str: Hex encoded.
  """
  return _checksum_file(hashlib.md5, path)


def sha256(data):
  """
  Return the sha256 of "data".

  Arguments:
      data (bytes): Data.

  Returns:
      str: Hex encoded.
  """
  return _checksum(hashlib.sha256, data)


def sha256_str(string, encoding='utf-8'):
  """
  Return the sha256 of string "data".

  Arguments:
      string: String.

  Returns:
      str: Hex encoded.
  """
  return _checksum_str(hashlib.sha256, string, encoding=encoding)


def sha256_list(*elems):
  """
  Return the sha256 of all elements of a list.

  Arguments:
      *elems: List of stringifiable data.

  Returns:
      str: Hex encoded.
  """
  return _checksum_list(hashlib.sha256, *elems)


def sha256_file(path: typing.Union[str, pathlib.Path]):
  """
  Return the sha256 of file at "path".

  Arguments:
      path (str): Path to file

  Returns:
      str: Hex encoded.
  """
  return _checksum_file(hashlib.sha256, path)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from setuptools import setup
from setuptools import find_packages

packages = find_packages()

setup(
    name = "labm8",
    version = "2019.10.17",
    description = "Utility libraries for doing science",
    long_description = open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers = ["Development Status :: 4 - Beta", "Environment :: Console", "Intended Audience :: Developers", "License :: OSI Approved :: Apache Software License", "Programming Language :: Python :: 3.6", "Programming Language :: Python :: 3.7", "Programming Language :: Python :: 3.8"],
    keywords = "utility library bazel protobuf",
    url = "https://github.com/ChrisCummins/labm8",
    author = "Chris Cummins",
    author_email = "chrisc.101@gmail.com",
    license = "Apache License, Version 2.0",
    packages=packages,
    install_requires=["SQLAlchemy==1.3.10", "Send2Trash==1.5.0", "absl-py==0.7.0", "checksumdir==1.0.5", "cycler==0.10.0", "decorator==4.3.0", "grpcio==1.18.0", "humanize==0.5.1", "kiwisolver==1.0.1", "matplotlib==2.2.0rc1", "mysqlclient==1.4.2.post1", "networkx==2.2", "numpy==1.16.4", "pandas==0.24.1", "protobuf==3.6.1", "py==1.5.2", "pyparsing==2.2.0", "python-dateutil==2.6.1", "pytz==2018.3", "scipy==1.2.1", "six==1.11.0"],
    zip_safe=False,
)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Text utilities.
"""
from __future__ import division

import re
import typing

import networkx as nx


class Error(Exception):
  """
  Module-level error.
  """
  pass


class TruncateError(Error):
  """
  Thrown in case of truncation error.
  """
  pass


def get_substring_idxs(substr, string):
  """
  Return a list of indexes of substr. If substr not found, list is
  empty.

  Arguments:
      substr (str): Substring to match.
      string (str): String to match in.

  Returns:
      list of int: Start indices of substr.
  """
  return [match.start() for match in re.finditer(substr, string)]


def truncate(string, maxchar):
  """
  Truncate a string to a maximum number of characters.

  If the string is longer than maxchar, then remove excess
  characters and append an ellipses.

  Arguments:

      string (str): String to truncate.
      maxchar (int): Maximum length of string in characters. Must be >= 4.

  Returns:

      str: Of length <= maxchar.

  Raises:

      TruncateError: In case of an error.
  """
  if maxchar < 4:
    raise TruncateError('Maxchar must be > 3')

  if len(string) <= maxchar:
    return string
  else:
    return string[:maxchar - 3] + '...'


def levenshtein(s1, s2):
  """
  Return the Levenshtein distance between two strings.

  Implementation of Levenshtein distance, one of a family of edit
  distance metrics.

  Based on: https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python

  Examples:

      >>> text.levensthein("foo", "foo")
      0

      >>> text.levensthein("foo", "fooo")
      1

      >>> text.levensthein("foo", "")
      3

      >>> text.levensthein("1234", "1 34")
      1

  Arguments:

      s1 (str): Argument A.
      s2 (str): Argument B.

  Returns:

      int: Levenshtein distance between the two strings.
  """
  # Left string must be >= right string.
  if len(s1) < len(s2):
    return levenshtein(s2, s1)

  # Distance is length of s1 if s2 is empty.
  if len(s2) == 0:
    return len(s1)

  previous_row = range(len(s2) + 1)
  for i, c1 in enumerate(s1):
    current_row = [i + 1]
    for j, c2 in enumerate(s2):
      insertions = previous_row[j + 1] + 1
      deletions = current_row[j] + 1
      substitutions = previous_row[j] + (c1 != c2)
      current_row.append(min(insertions, deletions, substitutions))
    previous_row = current_row

  return previous_row[-1]


def diff(s1, s2):
  """
  Return a normalised Levenshtein distance between two strings.

  Distance is normalised by dividing the Levenshtein distance of the
  two strings by the max(len(s1), len(s2)).

  Examples:

      >>> text.diff("foo", "foo")
      0

      >>> text.diff("foo", "fooo")
      1

      >>> text.diff("foo", "")
      1

      >>> text.diff("1234", "1 34")
      1

  Arguments:

      s1 (str): Argument A.
      s2 (str): Argument B.

  Returns:

      float: Normalised distance between the two strings.
  """
  return levenshtein(s1, s2) / max(len(s1), len(s2))


def AddWordToPrefixTree(trie: nx.DiGraph, word: str) -> None:
  """Add the given word to a prefix tree.

  Args:
    trie: The prefix tree.
    word: The word to add to the prefix tree.
  """
  current_node = 0
  for char in word:
    for neighbour_id in trie[current_node]:
      if trie.nodes[neighbour_id]['char'] == char:
        current_node = neighbour_id
        break
    else:
      new_node_id = max(trie.nodes) + 1
      trie.add_node(new_node_id, char=char)
      trie.add_edge(current_node, new_node_id)
      current_node = new_node_id

  trie.nodes[current_node]['word'] = word


def BuildPrefixTree(words: typing.Set[str]):
  """Construct a prefix tree from the given words.

  A prefix tree can be used for providing autocomplete results for a prefix,
  see AutoCompletePrefix(). This implementation uses a networkx directed graph
  to represent the tree, with a 'char' attribute attached to nodes, and a 'word'
  attribute containing whole-word sequences on terminal nodes.

  This implementation favors clarity over efficiency.

  Args:
    words: The words to construct a prefix tree from.

  Returns:
    A prefix tree.
  """
  trie = nx.DiGraph()
  trie.add_node(0)
  for word in words:
    AddWordToPrefixTree(trie, word)
  return trie


def PrefixTreeWords(trie: nx.DiGraph, root_node: int = 0) -> typing.List[str]:
  """Return all words in prefix tree.

  Args:
    trie: A prefix tree, constructed using BuildPrefixTree().
    root_node: The root node to get words from.

  Returns:
    The set of words in the prefix tree.
  """
  ret = set()
  word = trie.nodes[root_node].get('word')
  if word:
    ret.add(word)
  for neighbor_node in trie[root_node]:
    ret = ret.union(PrefixTreeWords(trie, root_node=neighbor_node))
  return ret


def AutoCompletePrefix(prefix: str, trie: nx.DiGraph) -> typing.Set[str]:
  """Return all words in prefix tree which start with the given prefix.

  Args:
    prefix: The prefix to match.
    trie: A prefix tree, constructed using BuildPrefixTree().

  Returns:
    The set of words in the prefix tree that begin with the prefix.

  Raises:
    ValueError: If prefix is not given.
    KeyError: If the prefix is not found.
  """
  if not prefix:
    raise ValueError('Prefix cannot be empty')

  ret = set()

  current_node = 0
  for char in prefix:
    for neighbour_id in trie[current_node]:
      if trie.nodes[neighbour_id]['char'] == char:
        current_node = neighbour_id
        node_word = trie.nodes[current_node].get('word')
        if node_word:
          ret.add(node_word)
        break
    else:
      raise KeyError(f"Prefix not found: '{prefix}'")

  return ret.union(PrefixTreeWords(trie, root_node=current_node))


def CamelCapsToUnderscoreSeparated(camel_caps_str: str):
  components = re.findall('[A-Z][^A-Z]*', camel_caps_str)
  assert components
  return '_'.join(x.lower() for x in components)


def StripSingleLineComments(
    string: str,
    start_comment_re: str = '(#|//)',
) -> str:
  """Strip line comments from a string.

  Args:
    string: The string to strip the comments of.
    start_comment_re: The regular expression to match the start of a line
      comment. By default, this matches Bash-style '#' and C-style '//'
      comments.

  Returns:
    The string.
  """
  comment_re = re.compile(f'{start_comment_re}.*')
  lines = [comment_re.sub('', line) for line in string.split('\n')]
  return '\n'.join(lines)



"""A module for launching docker images from within python applications."""
import contextlib
import pathlib
import random
import subprocess
import typing

from labm8 import app
from labm8 import bazelutil
from labm8 import labtypes


def IsDockerContainer() -> bool:
  """Determine if running inside a docker container."""
  return pathlib.Path('/.dockerenv').is_file()


def _Docker(cmd: typing.List[str], timeout: int = 60):
  """Build a docker process invocation."""
  cmd = ['timeout', '-s9', str(timeout), 'docker'] + [str(s) for s in cmd]
  app.Log(2, '$ %s', ' '.join(cmd))
  return cmd


class DockerImageRunContext(object):
  """A transient context for running docker images."""

  def __init__(self, image_name: str):
    self.image_name = image_name

  def _CommandLineInvocation(
      self,
      args: typing.List[str],
      flags: typing.Dict[str, str],
      volumes: typing.Dict[typing.Union[str, pathlib.Path], str],
      timeout: int,
      entrypoint: typing.Optional[str],
  ) -> typing.List[str]:
    """Build the command line arguments to execute the requested command.

    Args:
      args: A list of string positional arguments to pass to the docker image.
      flags: A map of flag arguments. The keys are prefixed with '--' and
        concatenated with the keys to produce arguments. E.g. {"foo": 2}
        equates to ["--foo", "2"].
      volumes: A map of shared volumes, e.g. {"/tmp": "/foo"} equates to
        the argument "-v/tmp:/foo". It is the responsibility of the calling
        code to ensure that the host paths are accessible to docker, and have
        the appropriate permissions for the docker user to access / modify. See
        //labm8:dockerutil_test for an example.
      timeout: The number of seconds to allow the image to run for before being
        killed. Killed processes will exit with returncode 9.
      entrypoint: An optional entrypoint for the docker image, equivalent to
        ["--entrypoint", entrypoint].

    Returns:
      The command line as a list of strings.
    """
    entrypoint_args = ['--entrypoint', entrypoint] if entrypoint else []
    volume_args = [f'-v{src}:{dst}' for src, dst in (volumes or {}).items()]
    flags_args = labtypes.flatten(
        [[f'--{k}', str(v)] for k, v in (flags or {}).items()],)
    return _Docker(
        ['run'] + entrypoint_args + volume_args + [self.image_name] + args +
        flags_args,
        timeout,
    )

  def CheckCall(
      self,
      args: typing.List[str],
      flags: typing.Dict[str, str] = None,
      volumes: typing.Dict[typing.Union[str, pathlib.Path], str] = None,
      timeout: int = 600,
      entrypoint: str = None,
  ) -> None:
    """Run the docker image with specified args.

    This attempts to emulate the behavior of subproces.check_call() for
    docker images.

    See _CommandLineInvocation() for details on args.
    """
    cmd = self._CommandLineInvocation(args, flags, volumes, timeout, entrypoint)
    subprocess.check_call(cmd)

  def CheckOutput(
      self,
      args: typing.List[str],
      flags: typing.Dict[str, str] = None,
      volumes: typing.Dict[typing.Union[str, pathlib.Path], str] = None,
      timeout: int = 600,
      entrypoint: str = None,
  ) -> str:
    """Run the docker image with specified args and return its output.

    This attempts to emulate the behavior of subproces.check_output() for
    docker images.

    See _CommandLineInvocation() for details on args.
    """
    cmd = self._CommandLineInvocation(args, flags, volumes, timeout, entrypoint)
    return subprocess.check_output(cmd, universal_newlines=True)


class BazelPy3Image(object):
  """A docker image created using bazel's py3_image() rule.

  To use a py3_image with this class, add the py3_image target with a ".tar"
  suffix as a data dependency to the bazel target, e.g.

      load("@io_bazel_rules_docker//python3:image.bzl", "py3_image")

      py3_image(
          name = "my_image",
          srcs = ["my_image.py"],
      )

      py_binary(
          name = "my_app",
          srcs = ["my_app.py"],
          data = [
              ":my_image.tar",
          ],
          deps = [
              "//labm8:app",
              "//labm8:dockerutil",
          ],
      )
  """

  def __init__(self, data_path: str):
    """Constructor.

    Args:
      path: The path to the data, including the name of the workspace.

    Raises:
      FileNotFoundError: If path is not a file.
    """
    super(BazelPy3Image, self).__init__()
    self.data_path = data_path
    self.tar_path = bazelutil.DataPath(f'phd/{data_path}.tar')

    components = self.data_path.split('/')
    self.image_name = f'bazel/{"/".join(components[:-1])}:{components[-1]}'

  def _TemporaryImageName(self) -> str:
    basename = self.data_path.split('/')[-1]
    random_suffix = ''.join(
        random.choice('0123456789abcdef') for _ in range(32))
    return f'phd_{basename}_tmp_{random_suffix}'

  @contextlib.contextmanager
  def RunContext(self) -> DockerImageRunContext:
    subprocess.check_call(
        _Docker(['load', '-i', str(self.tar_path)], timeout=600),)
    tmp_name = self._TemporaryImageName()
    subprocess.check_call(
        _Docker(['tag', self.image_name, tmp_name], timeout=60),)
    subprocess.check_call(_Docker(['rmi', self.image_name], timeout=60))
    yield DockerImageRunContext(tmp_name)
    # FIXME(cec): Using the --force flag here is almost certainly the wrong
    # thing, but I'm getting strange errors when trying to untag the image
    # otherwise:
    #   Error response from daemon: conflict: unable to remove repository
    #   reference "phd_..." (must force) - container ... is using its
    #   referenced image ...
    subprocess.check_call(_Docker(['rmi', '--force', tmp_name], timeout=60))



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility code for working with sqlalchemy."""
import time

import collections
import contextlib
import pathlib
import sqlalchemy as sql
import typing
from absl import flags as absl_flags
from sqlalchemy import func
from sqlalchemy import orm
from sqlalchemy.dialects import mysql
from sqlalchemy.ext import declarative

from labm8 import humanize
from labm8 import labdate
from labm8 import pbutil
from labm8 import text
from labm8.internal import logging

FLAGS = absl_flags.FLAGS

absl_flags.DEFINE_boolean(
    'sqlutil_echo',
    False,
    'If True, the Engine will log all statements as well as a repr() of their '
    'parameter lists to the engines logger, which defaults to sys.stdout.',
)
absl_flags.DEFINE_integer(
    'mysql_engine_pool_size',
    5,
    'The number of connections to keep open inside the connection pool. A '
    '--mysql_engine_pool_size of 0 indicates no limit',
)
absl_flags.DEFINE_integer(
    'mysql_engine_max_overflow',
    10,
    'The number of connections to allow in connection pool “overflow”, that '
    'is connections that can be opened above and beyond the '
    '--mysql_engine_pool_size setting',
)

# The Query type is returned by Session.query(). This is a convenience for type
# annotations.
Query = orm.query.Query


class DatabaseNotFound(FileNotFoundError):
  """An error that is raised if the requested database cannot be found."""

  def __init__(self, url: str):
    self._url = url

  @property
  def url(self):
    return self._url

  def __repr__(self) -> str:
    return f"Database not found: '{self.url}'"

  def __str__(self) -> str:
    return repr(self)


def Base(*args, **kwargs) -> sql.ext.declarative.DeclarativeMeta:
  """Construct a base class for declarative class definitions."""
  return sql.ext.declarative.declarative_base(*args, **kwargs)


def GetOrAdd(session: sql.orm.session.Session,
             model,
             defaults: typing.Dict[str, object] = None,
             **kwargs):
  """Instantiate a mapped database object.

  If the object is not in the database,
  add it. Note that no change is written to disk until commit() is called on the
  session.

  Args:
    session: The database session.
    model: The database table class.
    defaults: Default values for mapped objects.
    kwargs: The values for the table row.

  Returns:
    An instance of the model class, with the values specified.
  """
  instance = session.query(model).filter_by(**kwargs).first()
  if not instance:
    params = {
        k: v
        for k, v in kwargs.items()
        if not isinstance(v, sql.sql.expression.ClauseElement)
    }
    params.update(defaults or {})
    instance = model(**params)
    session.add(instance)
    logging.Log(
        logging.GetCallingModuleName(),
        5,
        'New record: %s(%s)',
        model.__name__,
        params,
    )
  return instance


def Get(session: sql.orm.session.Session,
        model,
        defaults: typing.Dict[str, object] = None,
        **kwargs):
  """Determine if a database object exists.

  Args:
    session: The database session.
    model: The database table class.
    defaults: Default values for mapped objects.
    kwargs: The values for the table row.

  Returns:
    An instance of the model class with the values specified, or None if the
    object is not in the database.
  """
  del defaults
  return session.query(model).filter_by(**kwargs).first()


def CreateEngine(url: str, must_exist: bool = False) -> sql.engine.Engine:
  """Create an sqlalchemy database engine.

  This is a convenience wrapper for creating an sqlalchemy engine, that also
  creates the database if required, and checks that the database exists. This
  means that it is less flexible than SqlAlchemy's create_engine() - only three
  combination of dialects and drivers are supported: sqlite, mysql, and
  postgresql.

  See https://docs.sqlalchemy.org/en/latest/core/engines.html for details.

  Additionally, this implements a custom 'file://' handler, which reads a URL
  from a local file, and returns a connection to the database addressed by the
  URL. Use this if you would like to keep sensitive information such as a MySQL
  database password out of your .bash_history.

  Examples:
    Create in-memory SQLite database:
    >>> engine = CreateEngine('sqlite://')

    Connect to an SQLite database at relative.db:
    >>> engine = CreateEngine('sqlite:///relative.db')

    Connect to an SQLite database at /absolute/path/to/db:
    >>> engine = CreateEngine('sqlite:////absolute/path/to/db')

    Connect to MySQL database:
    >>> engine = CreateEngine(
        'mysql://bob:password@localhost:1234/database?charset=utf8')

    Connect to PostgreSQL database:
    >>> engine.CreateEngine(
      'postgresql://bob:password@localhost:1234/database')

    Connect to a URL specified in the file /tmp/url.txt:
    >>> engine.CreateEngine('file:///tmp/url.txt')

    Connect to a URL specified in the file /tmp/url.txt, with the suffix
    '/database?charset=utf8':
    >>> engine.CreateEngine('file:///tmp/url.txt?/database?charset=utf8')

  Args:
    url: The URL of the database to connect to.
    must_exist: If True, raise DatabaseNotFound if it doesn't exist. Else,
        database is created if it doesn't exist.

  Returns:
    An SQLalchemy Engine instance.

  Raises:
    DatabaseNotFound: If the database does not exist and must_exist is set.
    ValueError: If the datastore backend is not supported.
  """
  engine_args = {}

  # Read and expand a `file://` prefixed URL.
  url = ExpandFileUrl(url)

  if url.startswith('mysql://'):
    # Support for MySQL dialect.

    # We create a throwaway engine that we use to check if the requested
    # database exists.
    engine = sql.create_engine('/'.join(url.split('/')[:-1]))
    database = url.split('/')[-1].split('?')[0]
    query = engine.execute(
        sql.text(
            'SELECT SCHEMA_NAME FROM '
            'INFORMATION_SCHEMA.SCHEMATA WHERE '
            'SCHEMA_NAME = :database',),
        database=database,
    )

    # Engine-specific options.
    engine_args['pool_size'] = FLAGS.mysql_engine_pool_size
    engine_args['max_overflow'] = FLAGS.mysql_engine_max_overflow

    if not query.first():
      if must_exist:
        raise DatabaseNotFound(url)
      else:
        # We can't use sql.text() escaping here because it uses single quotes
        # for escaping. MySQL only accepts backticks for quoting database
        # names.
        engine.execute(f'CREATE DATABASE `{database}`')
    engine.dispose()
  elif url.startswith('sqlite://'):
    # Support for SQLite dialect.

    # This project (phd) deliberately disallows relative paths due to Bazel
    # sandboxing.
    if url != 'sqlite://' and not url.startswith('sqlite:////'):
      raise ValueError('Relative path to SQLite database is not allowed')

    if url == 'sqlite://':
      if must_exist:
        raise ValueError(
            'must_exist=True not valid for in-memory SQLite database',)
    else:
      path = pathlib.Path(url[len('sqlite:///'):])
      if must_exist:
        if not path.is_file():
          raise DatabaseNotFound(url)
      else:
        # Make the parent directory for SQLite database if creating a new
        # database.
        path.parent.mkdir(parents=True, exist_ok=True)
  elif url.startswith('postgresql://'):
    # Support for PostgreSQL dialect.

    engine = sql.create_engine('/'.join(url.split('/')[:-1] + ['postgres']))
    conn = engine.connect()
    database = url.split('/')[-1]
    query = conn.execute(
        sql.text('SELECT 1 FROM pg_database WHERE datname = :database'),
        database=database,
    )
    if not query.first():
      if must_exist:
        raise DatabaseNotFound(url)
      else:
        # PostgreSQL does not let you create databases within a transaction, so
        # manually complete the transaction before creating the database.
        conn.execute(sql.text('COMMIT'))
        # PostgreSQL does not allow single quoting of database names.
        conn.execute(f'CREATE DATABASE {database}')
    conn.close()
    engine.dispose()
  else:
    raise ValueError(f"Unsupported database URL='{url}'")

  # Create the engine.
  engine = sql.create_engine(url,
                             encoding='utf-8',
                             echo=FLAGS.sqlutil_echo,
                             **engine_args)

  # Create and immediately close a connection. This is because SQLAlchemy engine
  # is lazily instantiated, so for connections such as SQLite, this line
  # actually creates the file.
  engine.connect().close()

  return engine


def ExpandFileUrl(url: str):
  """Expand URLs which begin with 'file://' by reading the file contents.

  If the URL does not begin with `file://`, it is returned unmodified.

  Args:
    url: The URL to expand, e.g. `file://path/to/file.txt?arg'

  Returns:
    The URL as interpreted by reading any URL file.

  Raises:
    ValueError: If the file path is invalid.
    FileNotFoundError: IF the file path does not exist.
  """
  if not url.startswith('file://'):
    return url

  # Split the URL into the file path, and the optional suffix.
  components = url.split('?')
  path, suffix = components[0], '?'.join(components[1:])

  # Strip the file:// prefix from the path.
  path = pathlib.Path(path[len('file://'):])

  if not path.is_absolute():
    raise ValueError('Relative path to file:// is not allowed')

  if not path.is_file():
    raise FileNotFoundError(f"File '{path}' not found")

  # Read the contents of the file, ignoring lines starting with '#'.
  with open(path) as f:
    file_url = '\n'.join(x for x in f.read().split('\n')
                         if not x.lstrip().startswith('#')).strip()

  # Append the suffix.
  file_url += suffix

  return file_url


def ColumnNames(model) -> typing.List[str]:
  """Return the names of all columns in a mapped object.

  Args:
    model: A mapped class.

  Returns:
    A list of string column names in the order that they are declared.
  """
  try:
    inst = sql.inspect(model)
    return [c_attr.key for c_attr in inst.mapper.column_attrs]
  except sql.exc.NoInspectionAvailable as e:
    raise TypeError(str(e))


class Session(orm.session.Session):
  """A subclass of the default SQLAlchemy Session with added functionality.

  An instance of this class is returned by Database.Session().
  """

  def GetOrAdd(self, model, defaults: typing.Dict[str, object] = None,
               **kwargs):
    """Instantiate a mapped database object.

    If the object is not in the database, add it. Note that no change is written
    to disk until commit() is called on the session.

    Args:
      model: The database table class.
      defaults: Default values for mapped objects.
      kwargs: The values for the table row.

    Returns:
      An instance of the model class, with the values specified.
    """
    return GetOrAdd(self, model, defaults, **kwargs)


class Database(object):
  """A base class for implementing databases."""

  SessionType = Session

  def __init__(self, url: str, declarative_base, must_exist: bool = False):
    """Instantiate a database object.

    Example:
      >>> db = Database('sqlite:////tmp/foo.db',
                        sqlalchemy.ext.declarative.declarative_base())

    Args:
      url: The URL of the database to connect to.
      declarative_base: The SQLAlchemy declarative base instance.
      must_exist: If True, raise DatabaseNotFound if it doesn't exist. Else,
        database is created if it doesn't exist.

    Raises:
      DatabaseNotFound: If the database does not exist and must_exist is set.
      ValueError: If the datastore backend is not supported.
    """
    self._url = url
    self.engine = CreateEngine(url, must_exist=must_exist)
    declarative_base.metadata.create_all(self.engine)
    declarative_base.metadata.bind = self.engine

    # Bind the Engine to a session maker, which instantiates our own Session
    # class, which is a subclass of the default SQLAlchemy Session with added
    # functionality.
    self.MakeSession = orm.sessionmaker(bind=self.engine, class_=Session)

  def Drop(self, are_you_sure_about_this_flag: bool = False):
    """Drop the database, irreverisbly destroying it.

    Be careful with this! After calling this method an a Database instance, no
    further operations can be made on it, and any Sessions should be discarded.

    Args:
      are_you_sure_about_this_flag: You should be sure.

    Raises:
      ValueError: In case you're not 100% sure.
    """
    if not are_you_sure_about_this_flag:
      raise ValueError("Let's take a minute to think things over")

    if self.url.startswith('mysql://'):
      engine = sql.create_engine('/'.join(self.url.split('/')[:-1]))
      database = self.url.split('/')[-1].split('?')[0]
      engine.execute(
          sql.text('DROP DATABASE IF EXISTS :database'),
          database=database,
      )
    elif self.url.startswith('sqlite://'):
      path = pathlib.Path(self.url[len('sqlite:///'):])
      assert path.is_file()
      path.unlink()
    else:
      raise NotImplementedError(
          "Unsupported operation DROP for database: '{self.url}'",)

  @property
  def url(self) -> str:
    """Return the URL of the database."""
    return self._url

  @contextlib.contextmanager
  def Session(self, commit: bool = False) -> Session:
    """Provide a transactional scope around a session.

    Args:
      commit: If true, commit session at the end of scope.

    Returns:
      A database session.
    """
    session = self.MakeSession()
    try:
      yield session
      if commit:
        session.commit()
    except:
      session.rollback()
      raise
    finally:
      session.close()

  @property
  def Random(self):
    """Get the backend-specific random function.

    This can be used to select a random row from a table, e.g.
        session.query(Table).order_by(db.Random()).first()
    """
    if self.url.startswith('mysql'):
      return func.rand
    else:
      return func.random  # for PostgreSQL, SQLite

  def __repr__(self) -> str:
    return self.url


class TablenameFromClassNameMixin(object):
  """A class mixin which derives __tablename__ from the class name.

  Add this mixin to a mapped table class to automatically set the set the
  __tablename__ property of a class to the lowercase name of the Python class.
  """

  @declarative.declared_attr
  def __tablename__(self):
    return self.__name__.lower()


class TablenameFromCamelCapsClassNameMixin(object):
  """A class mixin which derives __tablename__ from the class name.

  Add this mixin to a mapped table class to automatically set the set the
  __tablename__ property of a class to the name of the Python class with camel
  caps converted to underscores, e.g.

    class FooBar -> table "foo_bar".
  """

  @declarative.declared_attr
  def __tablename__(self):
    return text.CamelCapsToUnderscoreSeparated(self.__name__)


class PluralTablenameFromCamelCapsClassNameMixin(object):
  """A class mixin which derives __tablename__ from the class name.

  Add this mixin to a mapped table class to automatically set the set the
  __tablename__ property of a class to the pluralized name of the Python class
  with camel caps converted to underscores, e.g.

    class FooBar -> table "foo_bars".
  """

  @declarative.declared_attr
  def __tablename__(self):
    pluralised = humanize.Plural(2, self.__name__)
    pluralised = ' '.join(pluralised.split()[1:])
    return text.CamelCapsToUnderscoreSeparated(pluralised)


class ProtoBackedMixin(object):
  """A database table backed by protocol buffers.

  This class provides the abstract interface for sqlalchemy table classes which
  support serialization to and from protocol buffers.

  This is only an interface - inheriting classes must still inherit from
  sqlalchemy.ext.declarative.declarative_base().
  """
  proto_t = None

  def SetProto(self, proto: pbutil.ProtocolBuffer) -> None:
    """Set the fields of a protocol buffer with the values from the instance.

    Args:
      proto: A protocol buffer.
    """
    raise NotImplementedError(
        f'{type(self).__name__}.SetProto() not implemented',)

  def ToProto(self) -> pbutil.ProtocolBuffer:
    """Serialize the instance to protocol buffer.

    Returns:
      A protocol buffer.
    """
    proto = self.proto_t()
    self.SetProto(proto)
    return proto

  @classmethod
  def FromProto(
      cls,
      proto: pbutil.ProtocolBuffer,
  ) -> typing.Dict[str, typing.Any]:
    """Return a dictionary of instance constructor args from proto.

    Examples:
      Construct a table instance from proto:
      >>> table = Table(**Table.FromProto(proto))

      Construct a table instance and add to session:
      >>> session.GetOrAdd(Table, **Table.FromProto(proto))

    Args:
      proto: A protocol buffer.

    Returns:
      A dictionary of constructor arguments.
    """
    raise NotImplementedError(
        f'{type(self).__name__}.FromProto() not implemented',)

  @classmethod
  def FromFile(cls, path: pathlib.Path) -> typing.Dict[str, typing.Any]:
    """Return a dictionary of instance constructor args from proto file.

    Examples:
      Construct a table instance from proto file:
      >>> table = Table(**Table.FromFile(path))

      Construct a table instance and add to session:
      >>> session.GetOrAdd(Table, **Table.FromFile(path))

    Args:
      path: Path to a proto file.

    Returns:
      An instance.
    """
    proto = pbutil.FromFile(path, cls.proto_t())
    return cls.FromProto(proto)


# The results of an offset-limit batched query. The batch num is the current
# batch number. The offset is the offset into the results set, the limit is the
# last row in the results set, max_rows is the total number of rows in the query
# (only set if compute_max_rows, else None), and rows it the results.
OffsetLimitQueryResultsBatch = collections.namedtuple(
    'QueryResultsBatch',
    ['batch_num', 'offset', 'limit', 'max_rows', 'rows'],
)


def OffsetLimitBatchedQuery(
    query: Query,
    batch_size: int = 1000,
    start_at: int = 0,
    compute_max_rows: bool = False,
) -> typing.Iterator[OffsetLimitQueryResultsBatch]:
  """Split and return the rows resulting from a query in to batches.

  This iteratively runs the query `SELECT * FROM * OFFSET i LIMIT batch_size;`
  with `i` initialized to `start_at` and increasing by `batch_size` per
  iteration. Iteration terminates when the query returns no rows.

  This function is useful for returning row sets from enormous tables, where
  loading the full query results in to memory would take prohibitive time or
  resources.

  Args:
    query: The query to run.
    batch_size: The number of rows to return per batch.
    start_at: The initial offset into the table.
    compute_max_rows: If true

  Returns:
    A generator of OffsetLimitQueryResultsBatch tuples, where each tuple
    contains between 1 <= x <= `batch_size` rows.
  """
  max_rows = None
  if compute_max_rows:
    max_rows = query.count()

  batch_num = 0
  i = start_at
  while True:
    batch_num += 1
    batch = query.offset(i).limit(batch_size).all()
    if batch:
      yield OffsetLimitQueryResultsBatch(
          batch_num=batch_num,
          offset=i,
          limit=i + batch_size,
          max_rows=max_rows,
          rows=batch,
      )
      i += len(batch)
    else:
      break


class ColumnTypes(object):
  """Abstract class containing methods for generating column types."""

  def __init__(self):
    raise TypeError('abstract class')

  @staticmethod
  def BinaryArray(length: int):
    """Return a fixed size binary array column type.

    Args:
      length: The length of the column.

    Returns:
      A column type.
    """
    return sql.Binary(length).with_variant(mysql.BINARY(length), 'mysql')

  @staticmethod
  def LargeBinary():
    """Return a fixed size binary array column type.

    Returns:
      A column type.
    """
    return sql.LargeBinary().with_variant(sql.LargeBinary(2**31), 'mysql')

  @staticmethod
  def UnboundedUnicodeText():
    """Return an unbounded unicode text column type.

    This isn't truly unbounded, but 2^32 chars should be enough!

    Returns:
      A column type.
    """
    return sql.UnicodeText().with_variant(sql.UnicodeText(2**31), 'mysql')

  @staticmethod
  def IndexableString(length: int = None):
    """Return a string that is short enough that it can be used as an index.

    Returns:
      A column type.
    """
    # MySQL InnoDB tables use a default index key prefix length limit of 767.
    # https://dev.mysql.com/doc/refman/5.6/en/innodb-restrictions.html
    MAX_LENGTH = 767
    if length and length > MAX_LENGTH:
      raise ValueError(
          f'IndexableString requested length {length} is greater '
          f'than maximum allowed {MAX_LENGTH}',)
    return sql.String(MAX_LENGTH)

  @staticmethod
  def MillisecondDatetime():
    """Return a datetime type with millisecond precision.

    Returns:
      A column type.
    """
    return sql.DateTime().with_variant(mysql.DATETIME(fsp=3), 'mysql')


class ColumnFactory(object):
  """Abstract class containing methods for generating columns."""

  @staticmethod
  def MillisecondDatetime(
      nullable: bool = False,
      default=labdate.GetUtcMillisecondsNow,
  ):
    """Return a datetime column with millisecond precision.

    Returns:
      A column which defaults to UTC now.
    """
    return sql.Column(
        sql.DateTime().with_variant(
            mysql.DATETIME(fsp=3),
            'mysql',
        ),
        nullable=nullable,
        default=default,
    )


def ResilientAddManyAndCommit(db: Database, mapped: typing.Iterable[Base]):
  """Attempt to commit all mapped objects and return those that fail.

  This method creates a session and commits the given mapped objects.
  In case of error, this method will recurse up to O(log(n)) times, committing
  as many objects that can be as possible.

  Args:
    db: The database to add the objects to.
    mapped: A sequence of objects to commit.

  Returns:
    Any items in `mapped` which could not be committed, if any. Relative order
    of items is preserved.
  """
  failures = []

  if not mapped:
    return failures

  mapped = list(mapped)
  try:
    with db.Session(commit=True) as session:
      session.add_all(mapped)
  except sql.exc.SQLAlchemyError as e:
    logging.Log(
        logging.GetCallingModuleName(),
        1,
        'Caught error while committing %d mapped objects: %s',
        len(mapped),
        e,
    )

    # Divide and conquer. If we're committing only a single object, then a
    # failure to commit it means that we can do nothing other than return it.
    # Else, divide the mapped objects in half and attempt to commit as many of
    # them as possible.
    if len(mapped) == 1:
      return mapped
    else:
      mid = int(len(mapped) / 2)
      left = mapped[:mid]
      right = mapped[mid:]
      failures += ResilientAddManyAndCommit(db, left)
      failures += ResilientAddManyAndCommit(db, right)

  return failures


class BufferedDatabaseWriter(object):
  """A buffer for adding objects to a session with frequent commits.

  Use this class for cases when you are producing lots of mapped objects that
  you would like to commit to a database, but don't require them to be committed
  immediately. By buffering objects and committing them in batches, this class
  minimises the number of SQL statements that are executed, and is faster than
  creating and committing a session for every object.

  The Flush() method commits the contents of the buffer. The user is responsible
  for calling Flush() once the object reaches the end of its use. Alternatively,
  the Session() method creates a context which automatically calls Flush() at
  the end of its scope.

  Example usage:

    with BufferedDatabaseWriter(db).Session() as writer:
      for chunk in chunks_to_process:
        objs = ProcessChunk(chunk)
        writer.AddMany(objs)
  """

  def __init__(self, db: Database, flush_secs: int = 30, max_queue: int = 1024):
    """Create a BufferedDatabaseWriter.

    Args:
      db: The Database instance that this writer will add to.
      flush_secs: The number of seconds between commits.
      max_queue: The maximum size of the buffer between commits.
    """
    self._db = db
    self._last_commit = time.time()
    self._to_commit = []
    self._flush_secs = flush_secs
    self._max_queue = max_queue

  def __del__(self):
    self.Flush()

  @contextlib.contextmanager
  def Session(self) -> 'BufferedDatabaseWriter':
    """Yields a context manager which calls Flush() at the end of the scope.

    Returns:
      The `self` instance.
    """
    try:
      yield self
    finally:
      self.Flush()

  def AddOne(self, mapped: Base) -> None:
    """Record a mapped object."""
    self._to_commit.append(mapped)
    self.MaybeFlush()

  def AddMany(self, objects: typing.List[Base]) -> None:
    """Record multiple mapped objects."""
    self._to_commit += objects
    self.MaybeFlush()

  def MaybeFlush(self) -> None:
    """Determine if the buffer should be flushed, and if so, flush it."""
    if (len(self._to_commit) > self._max_queue or
        (time.time() - self._last_commit) > self._flush_secs):
      self.Flush()

  def Flush(self) -> None:
    """Commit all buffered mapped objects to database."""
    failures = ResilientAddManyAndCommit(self._db, self._to_commit)
    if len(failures):
      logging.Log(
          logging.GetCallingModuleName(),
          1,
          'BufferedDatabaseWriter failed to commit %d objects',
          len(failures),
      )
    self._to_commit = []
    self._last_commit = time.time()



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Profiling API for timing critical paths in code.
"""
import contextlib
import csv
import inspect
import os
import pathlib
import sys
import time
import typing

from labm8 import app
from labm8 import humanize
from labm8 import labdate
from labm8 import labtypes
from labm8 import system

_TIMERS = {}


def is_enabled():
  return os.environ.get('PROFILE') is not None


def enable():
  os.environ['PROFILE'] = '1'


def disable():
  os.environ.pop('PROFILE', None)


def isrunning(name):
  """
  Check if a timer is running.

  Arguments:

      name (str, optional): The name of the timer to check.

  Returns:

      bool: True if timer is running, else False.
  """
  return name in _TIMERS


def start(name):
  """
  Start a new profiling timer.

  Arguments:

      name (str, optional): The name of the timer to create. If no
        name is given, the resulting timer is anonymous. There can
        only be one anonymous timer.
      unique (bool, optional): If true, then ensure that timer name
        is unique. This is to prevent accidentally resetting an
        existing timer.

  Returns:

      bool: Whether or not profiling is enabled.
  """
  if is_enabled():
    _TIMERS[name] = time.time()
  return is_enabled()


def stop(name, file=sys.stderr):
  """
  Stop a profiling timer.

  Arguments:

      name (str): The name of the timer to stop. If no name is given, stop
          the global anonymous timer.

  Returns:

      bool: Whether or not profiling is enabled.

  Raises:

      KeyError: If the named timer does not exist.
  """
  if is_enabled():
    elapsed = (time.time() - _TIMERS[name])
    if elapsed > 60:
      elapsed_str = '{:.1f} m'.format(elapsed / 60)
    elif elapsed > 1:
      elapsed_str = '{:.1f} s'.format(elapsed)
    else:
      elapsed_str = '{:.1f} ms'.format(elapsed * 1000)

    del _TIMERS[name]
    print('[prof]', name, elapsed_str, file=file)
  return is_enabled()


def profile(fun, *args, **kwargs):
  """
  Profile a function.
  """
  timer_name = kwargs.pop('prof_name', None)

  if not timer_name:
    module = inspect.getmodule(fun)
    c = [module.__name__]
    parentclass = labtypes.get_class_that_defined_method(fun)
    if parentclass:
      c.append(parentclass.__name__)
    c.append(fun.__name__)
    timer_name = '.'.join(c)

  start(timer_name)
  ret = fun(*args, **kwargs)
  stop(timer_name)
  return ret


def timers():
  """
  Iterate over all timers.

  Returns:
      Iterable[str]: An iterator over all time names.
  """
  for name in _TIMERS:
    yield name


@contextlib.contextmanager
def Profile(
    name: typing.Union[str, typing.Callable[[int], str]] = '',
    print_to: typing.Callable[[str], None] = lambda msg: app.Log(1, msg),
):
  """A context manager which prints the elapsed time upon exit.

  Args:
    name: The name of the task being profiled. A callback may be provided which
      is called at task completion with the elapsed duration in seconds as its
      argument.
    print_to: The function to print the result to.
  """
  name = name or 'completed'
  start_time = time.time()
  yield
  elapsed = time.time() - start_time
  if callable(name):
    name = name(elapsed)
  print_to(f'{name} in {humanize.Duration(elapsed)}')


@contextlib.contextmanager
def ProfileToFile(file_object, name: str = ''):
  """A context manager which prints profiling output to a file.

  Args:
    file_object: A file object to write
    name: The name of the task being profiled.
  """

  def _WriteToFile(message: str):
    """Print message to file, appending a newline."""
    file_object.write(f'{message}\n')

  yield Profile(name=name, print_to=_WriteToFile)


class ProfilingEvent(object):

  def __init__(self, start_time: int, name: str):
    self.start_time = start_time
    self.name = name


@contextlib.contextmanager
def ProfileToStdout(name: str = ''):
  """A context manager which prints the elapsed time to stdout on exit.

  Args:
    name: The name of the task being profiled.
  """
  with Profile(name, print_to=print):
    yield


class AutoCsvProfiler(object):

  def __init__(self, directory: pathlib.Path, name: str = 'profile'):
    self._directory = pathlib.Path(directory)
    if not self._directory.is_dir():
      raise ValueError(f'Directory not found: {directory}')
    self._name = name

    # Create the name of the logfile now, so that is timestamped to the start of
    # execution.
    timestamp = labdate.MillisecondsTimestamp()
    log_name = '.'.join([self._name, system.HOSTNAME, str(timestamp), 'csv'])
    self._path = self._directory / log_name

    with self._writer() as writer:
      writer.writerow(
          ('Start Time (ms since UNIX epoch)', 'Elapsed Time (ms)', 'Event'),)

  @contextlib.contextmanager
  def Profile(self, event_name: str = ''):
    """A context manager which prints the elapsed time upon exit.

    Args:
      event_name: The name of the event being profiled.
    """
    event = ProfilingEvent(labdate.MillisecondsTimestamp(), event_name)
    yield event
    elapsed = labdate.MillisecondsTimestamp() - event.start_time
    with self._writer() as writer:
      writer.writerow((event.start_time, elapsed, event.name))

  @contextlib.contextmanager
  def _writer(self):
    with open(self.path, 'a') as f:
      writer = csv.writer(f)
      yield writer

  @property
  def path(self) -> pathlib.Path:
    return self._path



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lightweight routines for producing more friendly output.

Usage examples:

  'New messages: %s' % humanize.Commas(star_count)
    -> 'New messages: 58,192'

  'Found %s.' % humanize.Plural(error_count, 'error')
    -> 'Found 2 errors.'

  'Found %s.' % humanize.Plural(error_count, 'ox', 'oxen')
    -> 'Found 2 oxen.'

  'Copied at %s.' % humanize.DecimalPrefix(rate, 'bps', precision=3)
    -> 'Copied at 42.6 Mbps.'

  'Free RAM: %s' % humanize.BinaryPrefix(bytes_free, 'B')
    -> 'Free RAM: 742 MiB'

  'Finished all tasks in %s.' % humanize.Duration(elapsed_time)
    -> 'Finished all tasks in 34m 5s.'

These libraries are not a substitute for full localization.  If you
need localization, then you will have to think about translating
strings, formatting numbers in different ways, and so on.  Use
ICU if your application is user-facing.  Use these libraries if
your application is an English-only internal tool, and you are
tired of seeing "1 results" or "3450134804 bytes used".

Compare humanize.*Prefix() to C++ utilites HumanReadableNumBytes and
HumanReadableInt in strings/human_readable.h.
"""
# Adapted from Google's humanize module. Apache 2.0 License.
# Copyright 2008 Google Inc. All Rights Reserved.
# <https://github.com/google/google-apputils>
import datetime
import math
import re
from typing import Optional
from typing import Union

import humanize as humanize_lib

SIBILANT_ENDINGS = frozenset(['sh', 'ss', 'tch', 'ax', 'ix', 'ex'])
DIGIT_SPLITTER = re.compile(r'\d+|\D+').findall

# These are included because they are common technical terms.
SPECIAL_PLURALS = {
    'index': 'indices',
    'matrix': 'matrices',
    'vertex': 'vertices',
}

VOWELS = frozenset('AEIOUaeiou')


def Commas(value: Optional[int]):
  """Formats an integer with thousands-separating commas.

  Args:
    value: An integer.

  Returns:
    A string.
  """
  if value is None:
    return '0'
  if value < 0:
    sign = '-'
    value = -value
  else:
    sign = ''
  result = []
  while value >= 1000:
    result.append('%03d' % (value % 1000))
    value /= 1000
  result.append('%d' % value)
  return sign + ','.join(reversed(result))


def Plural(quantity, singular, plural=None):
  """Formats an integer and a string into a single pluralized string.

  Args:
    quantity: An integer.
    singular: A string, the singular form of a noun.
    plural: A string, the plural form.  If not specified, then simple
        English rules of regular pluralization will be used.

  Returns:
    A string.
  """
  return '%d %s' % (quantity, PluralWord(quantity, singular, plural))


def PluralWord(quantity, singular, plural=None):
  """Builds the plural of an English word.

  Args:
    quantity: An integer.
    singular: A string, the singular form of a noun.
    plural: A string, the plural form.  If not specified, then simple
        English rules of regular pluralization will be used.

  Returns:
    the plural form of the word.
  """
  if quantity == 1:
    return singular
  if plural:
    return plural
  if singular in SPECIAL_PLURALS:
    return SPECIAL_PLURALS[singular]

  # We need to guess what the English plural might be.  Keep this
  # function simple!  It doesn't need to know about every possibility;
  # only regular rules and the most common special cases.
  #
  # Reference: http://en.wikipedia.org/wiki/English_plural

  for ending in SIBILANT_ENDINGS:
    if singular.endswith(ending):
      return '%ses' % singular
  if singular.endswith('o') and singular[-2:-1] not in VOWELS:
    return '%ses' % singular
  if singular.endswith('y') and singular[-2:-1] not in VOWELS:
    return '%sies' % singular[:-1]
  return '%ss' % singular


def WordSeries(words, conjunction='and'):
  """Convert a list of words to an English-language word series.

  Args:
    words: A list of word strings.
    conjunction: A coordinating conjunction.

  Returns:
    A single string containing all the words in the list separated by commas,
    the coordinating conjunction, and a serial comma, as appropriate.
  """
  num_words = len(words)
  if num_words == 0:
    return ''
  elif num_words == 1:
    return words[0]
  elif num_words == 2:
    return (' %s ' % conjunction).join(words)
  else:
    return '%s, %s %s' % (', '.join(words[:-1]), conjunction, words[-1])


def AddIndefiniteArticle(noun):
  """Formats a noun with an appropriate indefinite article.

  Args:
    noun: A string representing a noun.

  Returns:
    A string containing noun prefixed with an indefinite article, e.g.,
    "a thing" or "an object".
  """
  if not noun:
    raise ValueError('argument must be a word: {!r}'.format(noun))
  if noun[0] in VOWELS:
    return 'an ' + noun
  else:
    return 'a ' + noun


def DecimalPrefix(quantity,
                  unit,
                  precision=1,
                  min_scale=0,
                  max_scale=None,
                  separator=' '):
  """Formats an integer and a unit into a string, using decimal prefixes.

  The unit will be prefixed with an appropriate multiplier such that
  the formatted integer is less than 1,000 (as long as the raw integer
  is less than 10**27).  For example:

    DecimalPrefix(576012, 'bps') -> '576 kbps'
    DecimalPrefix(576012, '') -> '576 k'
    DecimalPrefix(576, '') -> '576'
    DecimalPrefix(1574215, 'bps', 2) -> '1.6 Mbps'

  Only the SI prefixes which are powers of 10**3 will be used, so
  DecimalPrefix(100, 'thread') is '100 thread', not '1 hthread'.

  See also:
    BinaryPrefix()

  Args:
    quantity: A number.
    unit: A string, the dimension for quantity, with no multipliers (e.g.
        "bps").  If quantity is dimensionless, the empty string.
    precision: An integer, the minimum number of digits to display.
    min_scale: minimum power of 1000 to scale to, (None = unbounded).
    max_scale: maximum power of 1000 to scale to, (None = unbounded).

  Returns:
    A string, composed by the decimal (scaled) representation of quantity at the
    required precision, possibly followed by a space, the appropriate multiplier
    and the unit.
  """
  return _Prefix(
      quantity,
      unit,
      precision,
      separator,
      DecimalScale,
      min_scale=min_scale,
      max_scale=max_scale,
  )


def BinaryPrefix(quantity, unit, precision=1, separator=' '):
  """Formats an integer and a unit into a string, using binary prefixes.

  The unit will be prefixed with an appropriate multiplier such that
  the formatted integer is less than 1,024 (as long as the raw integer
  is less than 2**90).  For example:

    BinaryPrefix(576012, 'B') -> '562 KiB'
    BinaryPrefix(576012, '') -> '562 Ki'

  See also:
    DecimalPrefix()

  Args:
    quantity: A number.
    unit: A string, the dimension for quantity, with no multipliers (e.g.
        "B").  If quantity is dimensionless, the empty string.
    precision: An integer, the minimum number of digits to display.

  Returns:
    A string, composed by the decimal (scaled) representation of quantity at the
    required precision, possibly followed by a space, the appropriate multiplier
    and the unit.
  """
  return _Prefix(quantity, unit, precision, separator, BinaryScale)


def _Prefix(quantity, unit, precision, separator, scale_callable, **args):
  """Formats an integer and a unit into a string.

  Args:
    quantity: A number.
    unit: A string, the dimension for quantity, with no multipliers (e.g.
        "bps").  If quantity is dimensionless, the empty string.
    precision: An integer, the minimum number of digits to display.
    separator: The separator between quantity and unit.
    scale_callable: A callable, scales the number and units.
    **args: named arguments passed to scale_callable.

  Returns:
    A string.
  """
  if not quantity:
    return '0%s%s' % (separator, unit)

  if quantity in [float('inf'), float('-inf')] or math.isnan(quantity):
    return '%f%s%s' % (quantity, separator, unit)

  scaled_quantity, scaled_unit = scale_callable(quantity, unit, **args)

  if scaled_unit:
    separator = ' '

  print_pattern = '%%.%df%%s%%s' % max(
      0,
      (precision - int(math.log(abs(scaled_quantity), 10)) - 1),
  )

  return print_pattern % (scaled_quantity, separator, scaled_unit)


# Prefixes and corresponding min_scale and max_scale for decimal formating.
DECIMAL_PREFIXES = (
    'y',
    'z',
    'a',
    'f',
    'p',
    'n',
    u'µ',
    'm',
    '',
    'k',
    'M',
    'G',
    'T',
    'P',
    'E',
    'Z',
    'Y',
)
DECIMAL_MIN_SCALE = -8
DECIMAL_MAX_SCALE = 8


def DecimalScale(quantity, unit, min_scale=0, max_scale=None):
  """Get the scaled value and decimal prefixed unit in a tuple.

    DecimalScale(576012, 'bps') -> (576.012, 'kbps')
    DecimalScale(1574215, 'bps') -> (1.574215, 'Mbps')

  Args:
    quantity: A number.
    unit: A string.
    min_scale: minimum power of 1000 to normalize to (None = unbounded)
    max_scale: maximum power of 1000 to normalize to (None = unbounded)

  Returns:
    A tuple of a scaled quantity (float) and BinaryPrefix for the
    units (string).
  """
  if min_scale is None or min_scale < DECIMAL_MIN_SCALE:
    min_scale = DECIMAL_MIN_SCALE
  if max_scale is None or max_scale > DECIMAL_MAX_SCALE:
    max_scale = DECIMAL_MAX_SCALE
  powers = DECIMAL_PREFIXES[min_scale - DECIMAL_MIN_SCALE:max_scale -
                            DECIMAL_MIN_SCALE + 1]
  return _Scale(quantity, unit, 1000, powers, min_scale)


def BinaryScale(quantity, unit):
  """Get the scaled value and binary prefixed unit in a tupple.

    BinaryPrefix(576012, 'B') -> (562.51171875, 'KiB')

  Args:
    quantity: A number.
    unit: A string.

  Returns:
    A tuple of a scaled quantity (float) and BinaryPrefix for the
    units (string).
  """
  return _Scale(
      quantity,
      unit,
      1024,
      ('Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'),
  )


def _Scale(quantity, unit, multiplier, prefixes=None, min_scale=None):
  """Returns the formatted quantity and unit into a tuple.

  Args:
    quantity: A number.
    unit: A string
    multiplier: An integer, the ratio between prefixes.
    prefixes: A sequence of strings.
        If empty or None, no scaling is done.
    min_scale: minimum power of multiplier corresponding to the first prefix.
        If None assumes prefixes are for positive powers only.

  Returns:
    A tuple containing the raw scaled quantity (float) and the prefixed unit.
  """
  if (not prefixes or not quantity or math.isnan(quantity) or
      quantity in [float('inf'), float('-inf')]):
    return float(quantity), unit

  if min_scale is None:
    min_scale = 0
    prefixes = ('',) + tuple(prefixes)
  value, prefix = quantity, ''
  for power, prefix in enumerate(prefixes, min_scale):
    # This is more numerically accurate than '/ multiplier ** power'.
    value = float(quantity) * multiplier**-power
    if abs(value) < multiplier:
      break
  return value, prefix + unit


# Contains the fractions where the full range [1/n ... (n - 1) / n]
# is defined in Unicode.
FRACTIONS = {
    3: ('', u'⅓', u'⅔', ''),
    5: ('', u'⅕', u'⅖', u'⅗', u'⅘', ''),
    8: ('', u'⅛', u'¼', u'⅜', u'½', u'⅝', u'¾', u'⅞', ''),
}

FRACTION_ROUND_DOWN = 1.0 / (max(FRACTIONS.keys()) * 2.0)
FRACTION_ROUND_UP = 1.0 - FRACTION_ROUND_DOWN


def PrettyFraction(number, spacer=''):
  """Convert a number into a string that might include a unicode fraction.

  This method returns the integer representation followed by the closest
  fraction of a denominator 2, 3, 4, 5 or 8.
  For instance, 0.33 will be converted to 1/3.
  The resulting representation should be less than 1/16 off.

  Args:
    number: a python number
    spacer: an optional string to insert between the integer and the fraction
        default is an empty string.

  Returns:
    a unicode string representing the number.
  """
  # We do not want small negative numbers to display as -0.
  if number < -FRACTION_ROUND_DOWN:
    return u'-%s' % PrettyFraction(-number)
  number = abs(number)
  rounded = int(number)
  fract = number - rounded
  if fract >= FRACTION_ROUND_UP:
    return str(rounded + 1)
  errors_fractions = []
  for denominator, fraction_elements in FRACTIONS.items():
    numerator = int(round(denominator * fract))
    error = abs(fract - (float(numerator) / float(denominator)))
    errors_fractions.append((error, fraction_elements[numerator]))
  _, fraction_text = min(errors_fractions)
  if rounded and fraction_text:
    return u'%d%s%s' % (rounded, spacer, fraction_text)
  if rounded:
    return str(rounded)
  if fraction_text:
    return fraction_text
  return u'0'


def Duration(duration, separator=' '):
  """Formats a non-negative number of seconds into a human-readable string.

  Has nanosecond precision if passed a floating point number.

  Args:
    duration: A duration, either as number of seconds, or a timedelta.
    separator: A string separator between days, hours, minutes and seconds.

  Returns:
    Formatted string like '5d 12h 30m 45s'.
  """
  if isinstance(duration, datetime.timedelta):
    delta = duration
  else:
    try:
      delta = datetime.timedelta(seconds=int(duration))
    except OverflowError:
      return '>=' + TimeDelta(datetime.timedelta.max)
  delta_str = TimeDelta(delta, separator=separator)
  if isinstance(duration, float):
    value, unit = DecimalScale(duration % 1, 's', min_scale=-3)
    if value < 1 and unit == 'ns':
      # Special case for sub-nanosecond durations. If there's a second
      # component, discard the sub-nanosecond component. Else, report <1ns.
      if duration > 1:
        return delta_str
      else:
        return '<1ns'
    elif duration > 1:
      if unit != 's':
        # Append sub-second component to string.
        delta_str += f' {int(value)}{unit}'
    else:
      return f'{int(value)}{unit}'
  return delta_str


def LowPrecisionDuration(duration: Union[datetime.timedelta, int, float]):
  """Given a timedelta or a number of seconds, return a low precision, natural
  representation of the amount of time elapsed.

  This is similar to Duration(), but with a lower precision in the generated
  string, e.g. "a moment" rather than "530 ms".

  Args:
    duration: The duration to convert to a natural string.

  Returns:
    A natural representation of the duration.
  """
  return humanize_lib.naturaldelta(duration)


def TimeDelta(delta, separator=' '):
  """Format a datetime.timedelta into a human-readable string.

  Args:
    delta: The datetime.timedelta to format.
    separator: A string separator between days, hours, minutes and seconds.

  Returns:
    Formatted string like '5d 12h 30m 45s'.
  """
  parts = []
  seconds = delta.seconds
  if delta.days:
    parts.append('%dd' % delta.days)
  if seconds >= 3600:
    parts.append('%dh' % (seconds // 3600))
    seconds %= 3600
  if seconds >= 60:
    parts.append('%dm' % (seconds // 60))
    seconds %= 60
  seconds += delta.microseconds / 1e6
  if seconds or not parts:
    parts.append('%gs' % seconds)
  return separator.join(parts)


def Time(delta):
  return humanize_lib.naturaltime(delta)


def NaturalSortKey(data):
  """Key function for "natural sort" ordering.

  This key function results in a lexigraph sort. For example:
  - ['1, '3', '20'] (not ['1', '20', '3']).
  - ['Model 9', 'Model 70 SE', 'Model 70 SE2']
    (not ['Model 70 SE', 'Model 70 SE2', 'Model 9']).

  Usage:
    new_list = sorted(old_list, key=humanize.NaturalSortKey)
    or
    list_sort_in_place.sort(key=humanize.NaturalSortKey)

  Based on code by Steven Bazyl <sbazyl@google.com>.

  Args:
    data: str, The key being compared in a sort.

  Returns:
    A list which is comparable to other lists for the purpose of sorting.
  """

  class _StrComparableInt(object):
    """The original implementation relies on Python 2's broken str/int
    comparison. This class provides equivalent behavior."""

    def __init__(self, value):
      self._value = value

    def __lt__(self, rhs: '_StrComparableInt') -> bool:
      if isinstance(self._value, str) and isinstance(rhs._value, int):
        return False
      if isinstance(self._value, int) and isinstance(rhs._value, str):
        return True
      else:
        return self._value < rhs._value

    def __eq__(self, rhs) -> bool:
      return self._value == rhs._value

  segments = DIGIT_SPLITTER(data)
  for i, value in enumerate(segments):
    if value.isdigit():
      segments[i] = _StrComparableInt(int(value))
    else:
      segments[i] = _StrComparableInt(value)
  print('SEGMENTS', segments)
  return segments


def UnixTimestamp(unix_ts, tz):
  """Format a UNIX timestamp into a human-readable string.

  Args:
    unix_ts: UNIX timestamp (number of seconds since epoch). May be a floating
        point number.
    tz: datetime.tzinfo object, timezone to use when formatting. Typical uses
        might want to rely on datelib or pytz to provide the tzinfo object, e.g.
        use datelib.UTC, datelib.US_PACIFIC, or pytz.timezone('Europe/Dublin').

  Returns:
    Formatted string like '2013-11-17 11:08:27.720000 PST'.
  """
  date_time = datetime.datetime.fromtimestamp(unix_ts, tz)
  return date_time.strftime('%Y-%m-%d %H:%M:%S.%f %Z')


def AddOrdinalSuffix(value):
  """Adds an ordinal suffix to a non-negative integer (e.g. 1 -> '1st').

  Args:
    value: A non-negative integer.

  Returns:
    A string containing the integer with a two-letter ordinal suffix.
  """
  if value < 0 or value != int(value):
    raise ValueError('argument must be a non-negative integer: %s' % value)

  if value % 100 in (11, 12, 13):
    suffix = 'th'
  else:
    rem = value % 10
    if rem == 1:
      suffix = 'st'
    elif rem == 2:
      suffix = 'nd'
    elif rem == 3:
      suffix = 'rd'
    else:
      suffix = 'th'

  return str(value) + suffix



"""Implementations of custom flag types for absl.flags."""
import pathlib
from typing import Callable

from absl import flags as absl_flags


class PathParser(absl_flags.ArgumentParser):
  """Parser of path values."""

  def __init__(self,
               must_exist: bool = True,
               exist_ok: bool = True,
               is_dir: bool = False):
    """Create a path values parser.

    Args:
      must_exist: If true, the path mst exist.
      exist_ok: If not true, the path must not exist. Not effect if true.
        Implied true if must_exist.
      is_dir: If true, the path must be a directory.
    """
    self.must_exist = must_exist
    self.exist_ok = exist_ok
    self.is_dir = is_dir

  def parse(self, argument) -> pathlib.Path:
    """See base class."""
    val = self.convert(argument)
    if self.must_exist:
      if not val.exists():
        raise ValueError('not found')
      if self.is_dir and not val.is_dir():
        raise ValueError('not a directory')
      elif not self.is_dir and not val.is_file():
        raise ValueError('not a file')
    elif not self.exist_ok and val.exists():
      raise ValueError('already exists')
    return val

  def convert(self, argument: str) -> pathlib.Path:
    """Returns the value of this argument."""
    if not argument:
      raise TypeError('Path flag must be set')
    return pathlib.Path(argument)


class DatabaseParser(absl_flags.ArgumentParser):
  """Parser of path values."""

  def __init__(self, database_class, must_exist: bool = True):
    """Create a path values parser.

    Args:
      must_exist: If true, the database must exist. Else, it is created.
    """
    # TODO(cec): Raise TypeError if database_class is not a subclass of
    # 'sqlutil.Database'.
    self.database_class = database_class
    self.must_exist = must_exist

  def parse(self, argument) -> 'sqlutil.Database':
    """See base class."""
    return self.convert(argument)

  def convert(self, argument: str) -> Callable[[], 'sqlutil.Database']:
    """Returns the value of this argument."""
    if not argument:
      raise TypeError('Path flag must be set')
    return lambda: self.database_class(url=argument, must_exist=self.must_exist)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plotting and visualization utilities."""
import time
from scipy import stats

import matplotlib
import numpy as np
import pathlib
import seaborn as sns
import subprocess
import tempfile
import typing
from IPython.core import display
from matplotlib import axes
from matplotlib import pyplot as plt

from labm8 import app
from labm8 import fs


FLAGS = app.FLAGS


def Finalize(output: typing.Optional[typing.Union[str, pathlib.Path]] = None,
             figsize=None,
             tight=True,
             **savefig_opts):
  """Finalise a plot.

  Display or show the plot, then close it.

  Args:
    output: Path to save figure to. If not given, plot is shown.
    figsize: Figure size in inches.
    **savefig_opts: Any additional arguments to pass to
      plt.savefig(). Only required if output is not None.
  """
  # Set figure size.
  if figsize is not None:
    plt.gcf().set_size_inches(*figsize)

  # Set plot layout.
  if tight:
    plt.tight_layout()

  if output is None:
    plt.show()
  else:
    output = pathlib.Path(output)
    output.parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(str(output), **savefig_opts)
    app.Log(1, "Wrote '%s'", output)
  plt.close()


def ShowErrorBarCaps(ax: axes.Axes):
  """Show error bar caps.
  
  Seaborn paper style hides error bar caps. Call this function on an axes
  object to make them visible again.
  """
  for child in ax.get_children():
    if str(child).startswith('Line2D'):
      child.set_markeredgewidth(1)
      child.set_markersize(8)


def RotateXLabels(rotation: int = 90, ax: axes.Axes = None) -> None:
  """Rotate plot X labels anti-clockwise.

  Args:
    rotation: The number of degrees to rotate the labels by.
    ax: The plot axis.
  """
  ax = ax or plt.gca()
  plt.setp(ax.get_xticklabels(), rotation=rotation)


def RotateYLabels(rotation: int = 90, ax: axes.Axes = None):
  """Rotate plot Y labels anti-clockwise.

  Args:
    rotation: The number of degrees to rotate the labels by.
    ax: The plot axis.
  """
  ax = ax or plt.gca()
  plt.setp(ax.get_yticklabels(), rotation=rotation)


def FormatXLabelsAsTimestamps(
    format='%H:%M:%S',
    convert_to_seconds=lambda t: t / 1000,
    ax: axes.Axes = None,
) -> None:
  """Format the X labels

  Args:
    format: The time formatting string.
    convert_to_seconds: A function to convert values to seconds.
    ax: The plot axis.
  """
  ax = ax or plt.gca()
  formatter = matplotlib.ticker.FuncFormatter(
      lambda t, _: time.strftime(format, time.gmtime(convert_to_seconds(t))),
  )
  ax.xaxis.set_major_formatter(formatter)


def Distplot(
    x=None,
    hue=None,
    data=None,
    log_x: bool = False,
    log1p_x: bool = False,
    kde=False,
    bins=None,
    nbins: typing.Optional[int] = None,
    norm_hist=False,
    hue_order=None,
    ax=None,
):
  """An extension of seaborn distribution plots for grouped data.

  Args:
    x: The x label for dataframe.
    hue: An optional categorical grouping for data.
    data: Dataset for plotting.
    log_x: If True, log transform the data before plotting.
    log1p_x: If True, log1p transform the data before plotting.
    kde: Whether to plot a gaussian kernel density estimate.
    bins: Specification of hist bins, or None to use Freedman-Diaconis rule.
    nbins: If bins is None, specify the number of bins to use. If not provided,
      use min(30, Freedman-Diaconis rule).
    norm_hist: If True, the histogram height shows a density rather than a
      count. This is implied if a KDE or fitted density is plotted.
    hue_order: Order to plot the categorical levels in, otherwise the levels are
      inferred from the data objects.
    ax: If provided, plot on this axis.

  Returns:
    Returns the Axes object with the plot for further tweaking.
  """

  # Utility code taken from seaborn. See:
  # https://github.com/mwaskom/seaborn/blob/master/seaborn/distributions.py
  #
  # Copyright (c) 2012-2018, Michael L. Waskom
  # All rights reserved.
  #
  # Redistribution and use in source and binary forms, with or without
  # modification, are permitted provided that the following conditions are met:
  #
  # * Redistributions of source code must retain the above copyright notice,
  #   this list of conditions and the following disclaimer.
  #
  # * Redistributions in binary form must reproduce the above copyright notice,
  #   this list of conditions and the following disclaimer in the documentation
  #   and/or other materials provided with the distribution.
  #
  # * Neither the name of the project nor the names of its
  #   contributors may be used to endorse or promote products derived from
  #   this software without specific prior written permission.
  #
  # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  # AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  # IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  # ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
  # LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  # CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  # SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  # INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  # CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  # POSSIBILITY OF SUCH DAMAGE.

  def iqr(a):
    """Calculate the IQR for an array of numbers."""
    a = np.asarray(a)
    q1 = stats.scoreatpercentile(a, 25)
    q3 = stats.scoreatpercentile(a, 75)
    return q3 - q1

  def _freedman_diaconis_bins(a):
    """Calculate number of hist bins using Freedman-Diaconis rule."""
    # From https://stats.stackexchange.com/questions/798/
    a = np.asarray(a)
    if len(a) < 2:
      return 1
    h = 2 * iqr(a) / (len(a)**(1 / 3))
    # fall back to sqrt(a) bins if iqr is 0
    if h == 0:
      return int(np.sqrt(a.size))
    else:
      return int(np.ceil((a.max() - a.min()) / h))

  # End of seaborn utility code.

  values_to_plot = data[x]
  if log_x:
    values_to_plot = np.log(values_to_plot)
  elif log1p_x:
    values_to_plot = np.log1p(values_to_plot)

  if bins is None:
    if nbins is None:
      nbins = min(_freedman_diaconis_bins(values_to_plot), 50)
    bins = np.linspace(min(values_to_plot), max(values_to_plot), nbins)

  if ax is None:
    ax = plt.gca()

  if hue is None:
    sns.distplot(
        values_to_plot,
        kde=kde,
        bins=bins,
        label=x,
        norm_hist=norm_hist,
        ax=ax,
    )
  else:
    hue_order = hue_order or sorted(set(data[hue]))
    for h in hue_order:
      sns.distplot(
          values_to_plot[data[hue] == h],
          kde=kde,
          bins=bins,
          label=h,
          norm_hist=norm_hist,
          ax=ax,
      )
    plt.legend()

  return ax


def SummarizeFloats(floats: typing.Iterable[float], nplaces: int = 2) -> str:
  """Summarize a sequence of floats."""
  arr = np.array(list(floats), dtype=np.float32)
  percs = ' '.join([
      f'{p}%={np.percentile(arr, p):.{nplaces}f}' for p in [0, 50, 95, 99, 100]
  ])
  return (
      f'n={len(arr)}, mean={arr.mean():.{nplaces}f}, stdev={arr.std():.{nplaces}f}, '
      f'percentiles=[{percs}]')


def SummarizeInts(ints: typing.Iterable[int]) -> str:
  """Summarize a sequence of ints."""
  arr = np.array(list(ints), dtype=np.int32)
  percs = ' '.join(
      [f'{p}%={np.percentile(arr, p):.0f}' for p in [0, 50, 95, 99, 100]],)
  return (f'n={len(arr)}, mean={arr.mean():.2f}, stdev={arr.std():.2f}, '
          f'percentiles=[{percs}]')


def PlotDot(dot: str) -> None:
  """Compile and display the given dot plot."""
  with tempfile.TemporaryDirectory() as d:
    dot_path = pathlib.Path(d) / 'dot.dot'
    png_path = pathlib.Path(d) / 'dot.png'

    fs.Write(dot_path, dot.encode('utf-8'))
    try:
      subprocess.check_call(['dot', str(dot_path), '-Tpng', '-o', str(png_path)])
    except subprocess.CalledProcessError as e:
      raise ValueError(f"Failed to process dotgraph: {dot}")
    display.display(display.Image(filename=f'{d}/dot.png'))



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility code for working with Protocol Buffers."""
import collections
import gzip
import json
import pathlib
import subprocess
import typing

import google.protobuf.json_format
import google.protobuf.message
import google.protobuf.text_format

# A type alias for annotating methods which take or return protocol buffers.
ProtocolBuffer = typing.Any

# A type alias for protocol enum fields.
Enum = int


class ProtoValueError(ValueError):
  """Raised in case of a value error from a proto."""
  pass


class EncodeError(ProtoValueError):
  """Raised in case of error encoding a proto."""
  pass


class DecodeError(ProtoValueError):
  """Raised in case of error decoding a proto."""
  pass


class ProtoWorkerTimeoutError(subprocess.CalledProcessError):
  """Raised is a protobuf worker binary times out."""

  def __init__(
      self,
      cmd: typing.List[str],
      timeout_seconds: int,
      returncode: int,
  ):
    self.cmd = cmd
    self.timeout_seconds = timeout_seconds
    # subprocess.CalledProcessError.str() requires a returncode attribute.
    self.returncode = returncode

  def __repr__(self) -> str:
    return (f'Proto worker timeout after {self.timeout_seconds} '
            f"seconds: {' '.join(self.cmd)}")


def FromString(
    string: str,
    message: ProtocolBuffer,
    uninitialized_okay: bool = False,
) -> ProtocolBuffer:
  """Read a text format protocol buffer from a string.

  Args:
    string: A text format protocol buffer.
    message: A message instance to read into.
    uninitialized_okay: If True, do not require that decoded messages be
      initialized. If False, DecodeError is raised.

  Returns:
    The parsed message (same as the message argument).

  Raises:
    DecodeError: If the file cannot be decoded to the given message type, or if
      after decoding, the message is not initialized and uninitialized_okay is
      False.
  """
  try:
    google.protobuf.text_format.Merge(string, message)
  except google.protobuf.text_format.ParseError as e:
    raise DecodeError(e)

  if not uninitialized_okay and not message.IsInitialized():
    raise DecodeError(f'Required fields not set')

  return message


def FromFile(
    path: pathlib.Path,
    message: ProtocolBuffer,
    assume_filename: typing.Optional[typing.Union[str, pathlib.Path]] = None,
    uninitialized_okay: bool = False,
) -> ProtocolBuffer:
  """Read a protocol buffer from a file.

  This method uses attempts to guess the encoding from the path suffix,
  supporting binary, text, and json formatted messages. The mapping of suffixes
  to formatting is, in order:
      *.txt.gz: Gzipped text.
      *.txt: Text.
      *.pbtxt.gz: Gzipped text.
      *.pbtxt: Text.
      *.json.gz: Gzipped JSON.
      *.json: JSON.
      *.gz: Gzipped encoded string.
      *: Encoded string.

  Args:
    path: Path to the proto file.
    message: A message instance to read into.
    assume_filename: For the purpose of determining the encoding from the file
      extension, use this name rather than the true path.
    uninitialized_okay: If True, do not require that decoded messages be
      initialized. If False, DecodeError is raised.

  Returns:
    The parsed message (same as the message argument).

  Raises:
    FileNotFoundError: If the path does not exist.
    IsADirectoryError: If the path is a directory.
    DecodeError: If the file cannot be decoded to the given message type, or if
      after decoding, the message is not initialized and uninitialized_okay is
      False.
  """
  if not path.is_file():
    if path.is_dir():
      raise IsADirectoryError(f"Path is a directory: '{path}'")
    else:
      raise FileNotFoundError(f"File not found: '{path}'")

  suffixes = pathlib.Path(
      assume_filename,).suffixes if assume_filename else path.suffixes
  if suffixes and suffixes[-1] == '.gz':
    suffixes.pop()
    open_function = gzip.open
  else:
    open_function = open

  suffix = suffixes[-1] if suffixes else ''
  try:
    with open_function(path, 'rb') as f:
      if suffix == '.txt' or suffix == '.pbtxt':
        # Allow uninitialized fields here because we will catch the error later,
        # allowing us to report the path of the proto.
        FromString(f.read().decode('utf-8'), message, uninitialized_okay=True)
      elif suffix == '.json':
        google.protobuf.json_format.Parse(f.read(), message)
      else:
        message.ParseFromString(f.read())
  except (
      google.protobuf.text_format.ParseError,
      google.protobuf.json_format.ParseError,
  ) as e:
    # The exception raised during parsing depends on the message format. Catch
    # them all under a single DecodeError exception type.
    raise DecodeError(e)

  if not uninitialized_okay and not message.IsInitialized():
    raise DecodeError(f"Required fields not set: '{path}'")

  return message


def ToFile(
    message: ProtocolBuffer,
    path: pathlib.Path,
    exist_ok: bool = True,
    assume_filename: typing.Optional[typing.Union[str, pathlib.Path]] = None,
) -> ProtocolBuffer:
  """Write a protocol buffer to a file.

  This method uses attempts to guess the encoding from the path suffix,
  supporting binary, text, and json formatted messages. The mapping of suffixes
  to formatting is, in order:
      *.txt.gz: Gzipped text format.
      *.txt: Text format.
      *.pbtxt.gz: Gzipped text format.
      *.pbtxt: Text format.
      *.json.gz: Gzipped JSON format.
      *.json: JSON format.
      *.gz: Gzipped binary format.
      *: Binary format.

  Args:
    message: A message instance to write to file. The message must be
      initialized, i.e. have all required fields set.
    path: Path to the proto file.
    exist_ok: If True, overwrite existing file.
    assume_filename: For the purpose of determining the encoding from the file
      extension, use this name rather than the true path.

  Returns:
    The parsed message (same as the message argument).

  Raises:
    EncodeError: If the message is not initialized, i.e. it is missing required
      fields.
    FileNotFoundError: If the parent directory of the requested path does not
      exist.
    IsADirectoryError: If the requested path is a directory.
    FileExistsError: If the requested path already exists and exist_ok is False.
  """
  if not exist_ok and path.exists():
    raise FileExistsError(f'Refusing to overwrite {path}')

  # The SerializeToString() method refuses to encode a message which is not
  # initialized, whereas the MessageToString() and MessageToJson() methods do
  # not. This API should be consistent, so we enforce that all formats require
  # the message to be initialized.
  if not message.IsInitialized():
    class_name = type(message).__name__
    raise EncodeError(f"Required fields not set: '{class_name}'")

  suffixes = pathlib.Path(
      assume_filename,).suffixes if assume_filename else path.suffixes
  if suffixes and suffixes[-1] == '.gz':
    suffixes.pop()
    open_function = gzip.open
  else:
    open_function = open

  suffix = suffixes[-1] if suffixes else ''
  mode = 'wt' if suffix in {'.txt', '.pbtxt', '.json'} else 'wb'

  with open_function(path, mode) as f:
    if suffix == '.txt' or suffix == '.pbtxt':
      f.write(google.protobuf.text_format.MessageToString(message))
    elif suffix == '.json':
      f.write(
          google.protobuf.json_format.MessageToJson(
              message,
              preserving_proto_field_name=True,
          ),)
    else:
      f.write(message.SerializeToString())

  return message


def ToJson(message: ProtocolBuffer) -> 'jsonutil.JSON':
  """Return a JSON encoded representation of a protocol buffer.

  Args:
    message: The message to convert to JSON.

  Returns:
    JSON encoded message.
  """
  return google.protobuf.json_format.MessageToDict(
      message,
      preserving_proto_field_name=True,
  )


def _TruncatedString(string: str, n: int = 80) -> str:
  """Return the truncated first 'n' characters of a string.

  Args:
    string: The string to truncate.
    n: The maximum length of the string to return.

  Returns:
    The truncated string.
  """
  if len(string) > n:
    return string[:n - 3] + '...'
  else:
    return string


def _TruncateDictionaryStringValues(
    data: 'jsonutil.JSON',
    n: int = 62,
) -> 'jsonutil.JSON':
  """Truncate all string values in a nested dictionary.

  Args:
    data: A dictionary.

  Returns:
    The dictionary.
  """
  for key, value in data.items():
    if isinstance(value, collections.Mapping):
      data[key] = _TruncateDictionaryStringValues(data[key])
    elif isinstance(value, str):
      data[key] = _TruncatedString(value, n)
    else:
      data[key] = value
  return data


def PrettyPrintJson(message: ProtocolBuffer, truncate: int = 52) -> str:
  """Return a pretty printed JSON string representation of the message.

  Args:
    message: The message to pretty print.
    truncate: The length to truncate string values. Truncation is disabled if
      this argument is None.

  Returns:
    JSON string.
  """
  data = ToJson(message)
  return json.dumps(
      _TruncateDictionaryStringValues(data) if truncate else data,
      indent=2,
      sort_keys=True,
  )


def RaiseIfNotSet(
    proto: ProtocolBuffer,
    field: str,
    err: ValueError,
) -> typing.Any:
  """Check that a proto field is set before returning it.

  Args:
    proto: A message instance.
    field: The name of the field.
    err: The exception class to raise.

  Returns:
    The value of the field.

  Raises:
    ValueError: If the field is not set.
  """
  if not proto.HasField(field):
    raise err(f'datastore field {field} not set')
  elif not getattr(proto, field):
    raise err(f'datastore field {field} not set')
  return getattr(proto, field)


def ProtoIsReadable(
    path: typing.Union[str, pathlib.Path],
    message: ProtocolBuffer,
) -> bool:
  """Return whether a file is a readable protocol buffer.

  Arguments:
    path: The path of the file to read.
    message: An instance of the message type.

  Returns:
    True if contents of path can be parsed as an instance of message, else
    False.
  """
  try:
    FromFile(pathlib.Path(path), message)
    return True
  except:
    return False


def AssertFieldIsSet(
    proto: ProtocolBuffer,
    field_name: str,
    fail_message: str = None,
) -> typing.Optional[typing.Any]:
  """Assert that protocol buffer field is set.

  Args:
    proto: A proto message instance.
    field_name: The name of the field to assert the constraint on.
    fail_message: An optional message to raise the ProtoValueError
      with if the assertion fails. If not provided, a default message is used.

  Returns:
    The value of the field, if the field has a value. Even though a field may
      be set, it may not have a value. For example, if any of a 'oneof' fields
      is set, then this function will return True for the name of the oneof,
      but the return value will be None.

  Raises:
    ValueError: If the requested field does not exist in the proto schema.
    ProtoValueError: If the field is not set.
  """
  if not proto.HasField(field_name):
    proto_class_name = type(proto).__name__
    raise ProtoValueError(
        fail_message or f"Field not set: '{proto_class_name}.{field_name}'",)
  return getattr(proto, field_name) if hasattr(proto, field_name) else None


def AssertFieldConstraint(
    proto: ProtocolBuffer,
    field_name: str,
    constraint: typing.Callable[[typing.Any], bool] = lambda x: True,
    fail_message: str = None,
) -> typing.Any:
  """Assert a constraint on the value of a protocol buffer field.

  Args:
    proto: A proto message instance.
    field_name: The name of the field to assert the constraint on.
    constraint: A constraint checking function to call with the value of the
      field. The function must return True if the constraint check passes, else
      False. If no constraint is specified, this callback always returns True.
      This still allows you to use this function to check if a field is set.
    fail_message: An optional message to raise the ProtoValueError
      with if the assertion fails. If not provided, default messages are used.

  Returns:
    The value of the field.

  Raises:
    ValueError: If the requested field does not exist in the proto schema.
    ProtoValueError: If the field is not set, or if the constraint callback
      returns False for the field's value.
  """
  value = AssertFieldIsSet(proto, field_name, fail_message)
  if not constraint(value):
    proto_class_name = type(proto).__name__
    raise ProtoValueError(
        fail_message or
        f"Field fails constraint check: '{proto_class_name}.{field_name}'",)
  else:
    return value


def RunProcessMessage(
    cmd: typing.List[str],
    input_proto: ProtocolBuffer,
    timeout_seconds: int = 360,
    env: typing.Dict[str, str] = None,
) -> str:
  # Run the C++ worker process, capturing it's output.
  process = subprocess.Popen(
      ['timeout', '-s9', str(timeout_seconds)] + cmd,
      stdin=subprocess.PIPE,
      stdout=subprocess.PIPE,
      env=env,
  )
  # Send the input proto to the C++ worker process.
  # TODO: Add timeout.
  stdout, _ = process.communicate(input_proto.SerializeToString())

  # TODO: Check signal value, not hardcoded int.
  if process.returncode == -9 or process.returncode == 9:
    raise ProtoWorkerTimeoutError(
        cmd=cmd,
        timeout_seconds=timeout_seconds,
        returncode=process.returncode,
    )
  elif process.returncode:
    raise subprocess.CalledProcessError(process.returncode, cmd)

  return stdout


def RunProcessMessageToProto(
    cmd: typing.List[str],
    input_proto: ProtocolBuffer,
    output_proto: ProtocolBuffer,
    timeout_seconds: int = 360,
    env: typing.Dict[str, str] = None,
):
  stdout = RunProcessMessage(
      cmd,
      input_proto,
      timeout_seconds=timeout_seconds,
      env=env,
  )
  output_proto.ParseFromString(stdout)
  return output_proto


def RunProcessMessageInPlace(
    cmd: typing.List[str],
    input_proto: ProtocolBuffer,
    timeout_seconds: int = 360,
    env: typing.Dict[str, str] = None,
):
  input_proto.ParseFromString(
      RunProcessMessage(
          cmd,
          input_proto,
          timeout_seconds=timeout_seconds,
          env=env,
      ),)
  return input_proto


class ProtoBackedMixin(object):
  """A class backed by protocol buffers.

  This mixin provides the abstract interface for classes which support
  serialization of instances to and from protocol buffers.

  Inheriting classes must set the proto_t class attribute, and  implement the
  SetProto() and FromProto() methods.

  Attributes:
    proto_t: The protocol buffer class that backs instances of this class.
  """
  # Inheritinc classes must set this attribute to the Protocol Buffer class.
  proto_t = None

  def SetProto(self, proto: ProtocolBuffer) -> None:
    """Set the fields of a protocol buffer with the values of the instance.

    It is the responsibility of the inheriting class to ensure that all required
    instance variables are recorded as fields in this proto.

    Args:
      proto: A protocol buffer.
    """
    # ABSTRACT METHOD. Inheriting classes must implement!
    raise NotImplementedError(
        f'{type(self).__name__}.SetProto() not implemented',)

  @classmethod
  def FromProto(cls, proto: ProtocolBuffer) -> 'ProtoBackedMixin':
    """Return an instance of the class from proto.

    It is the responsibility of the inheriting class to ensure that all required
    instance variables are set according to the fields in the proto.

    Args:
      proto: A protocol buffer.

    Returns:
      An instance of the class.
    """
    # ABSTRACT METHOD. Inheriting classes must implement!
    raise NotImplementedError(
        f'{type(self).__name__}.FromProto() not implemented',)

  def ToProto(self) -> ProtocolBuffer:
    """Serialize the instance to protocol buffer.

    It is the responsibility of the inheriting class to set the proto_t class
    attribute to the

    Returns:
      A protocol buffer.
    """
    proto = self.proto_t()
    self.SetProto(proto)
    return proto

  @classmethod
  def FromProtoFile(cls, path: pathlib.Path) -> 'ProtoBackedMixin':
    """Return an instance of the class from serialized proto file.

    Args:
      path: Path to a proto file.

    Returns:
      An instance.
    """
    return cls.FromProto(FromFile(path, cls.proto_t()))



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python type utilities.
"""
import inspect
import itertools
import sys
import typing
from collections import Mapping

from six import string_types


def is_str(s):
  """
  Return whether variable is string type.

  On python 3, unicode encoding is *not* string type. On python 2, it is.

  Arguments:
      s: Value.

  Returns:
      bool: True if is string, else false.
  """
  return isinstance(s, string_types)


def is_dict(obj):
  """
  Check if an object is a dict.
  """
  return isinstance(obj, dict)


def is_seq(obj):
  """
  Check if an object is a sequence.
  """
  return (not is_str(obj) and not is_dict(obj) and
          (hasattr(obj, '__getitem__') or hasattr(obj, '__iter__')))


def flatten(lists):
  """
  Flatten a list of lists.
  """
  return [item for sublist in lists for item in sublist]


def update(dst, src):
  """
  Recursively update values in dst from src.

  Unlike the builtin dict.update() function, this method will descend into
  nested dicts, updating all nested values.

  Arguments:
      dst (dict): Destination dict.
      src (dict): Source dict.

  Returns:
      dict: dst updated with entries from src.
  """
  for k, v in src.items():
    if isinstance(v, Mapping):
      r = update(dst.get(k, {}), v)
      dst[k] = r
    else:
      dst[k] = src[k]
  return dst


def dict_values(src):
  """
  Recursively get values in dict.

  Unlike the builtin dict.values() function, this method will descend into
  nested dicts, returning all nested values.

  Arguments:
      src (dict): Source dict.

  Returns:
      list: List of values.
  """
  for v in src.values():
    if isinstance(v, dict):
      for v in dict_values(v):
        yield v
    else:
      yield v


def get_class_that_defined_method(meth):
  """
  Return the class that defines a method.

  Arguments:
      meth (str): Class method.

  Returns:
      class: Class object, or None if not a class method.
  """
  if sys.version_info >= (3, 0):
    # Written by @Yoel http://stackoverflow.com/a/25959545
    if inspect.ismethod(meth):
      for cls in inspect.getmro(meth.__self__.__class__):
        if cls.__dict__.get(meth.__name__) is meth:
          return cls
      meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
      cls = getattr(
          inspect.getmodule(meth),
          meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
      )
      if isinstance(cls, type):
        return cls
  else:
    try:
      # Writted by @Alex Martelli http://stackoverflow.com/a/961057
      for cls in inspect.getmro(meth.im_class):
        if meth.__name__ in cls.__dict__:
          return cls
    except AttributeError:
      return None
  return None


class ReprComparable(object):
  """
  An abstract class which may be inherited from in order to enable __repr__.
  """

  def __lt__(self, other):
    return str(self) < str(other)

  def __le__(self, other):
    return str(self) <= str(other)

  def __eq__(self, other):
    return str(self) == str(other)

  def __ne__(self, other):
    return str(self) != str(other)

  def __gt__(self, other):
    return str(self) > str(other)

  def __ge__(self, other):
    return str(self) >= str(other)


def PairwiseIterator(iterable: typing.Iterator[typing.Any],
                    ) -> typing.Iterator[typing.Tuple[typing.Any, typing.Any]]:
  """Construct a pairwise iterator for a input generator.

  Given an iterator, produces an iterator of overlapping pairs from the input:
  s -> (s0,s1), (s1,s2), (s2, s3), ...

  Args:
    iterable: The input iterable. Once called, this iterable should not be
      used any more.

  Returns:
    An iterator of pairs.
  """
  # Split the iterable into two.
  a, b = itertools.tee(iterable)
  # Advance the second iterable by one.
  next(b, None)
  # Return the pairs.
  return zip(a, b)


def SetDiff(
    a: typing.Iterator[typing.Any],
    b: typing.Iterator[typing.Any],
) -> typing.List[typing.Any]:
  """Return the set difference between two sequences.

  Args:
    a: An iterable.
    b: An iterable.

  Returns:
    The difference between the elements in the two iterables as a set.
  """
  set_a = set(a)
  set_b = set(b)
  return (set_a - set_b).union(set_b - set_a)


def AllSubclassesOfClass(cls: typing.Type) -> typing.Set[typing.Type]:
  """Return the set of subclasses of a base class.

  This recursively returns all nested subclasses of a base class.

  Example:
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(B): pass
    >>> AllSubclassesOfClass(A)
    {B, C}

  Args:
    cls: The class to return subclasses of.

  Returns:
    A set of class types.
  """
  return set(cls.__subclasses__()).union(
      [s for c in cls.__subclasses__() for s in AllSubclassesOfClass(c)],)


def Chunkify(iterable: typing.Iterable[typing.Any],
             chunk_size: int) -> typing.Iterable[typing.List[typing.Any]]:
  """Split an iterable into chunks of a given size.

  Args:
    iterable: The iterable to split into chunks.
    chunk_size: The size of the chunks to return.

  Returns:
    An iterator over chunks of the input iterable.
  """
  i = iter(iterable)
  piece = list(itertools.islice(i, chunk_size))
  while piece:
    yield piece
    piece = list(itertools.islice(i, chunk_size))
